"""
CCRS Constants | Cannlytics
Copyright (c) 2022-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 4/12/2022
Updated: 1/3/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Helpful CCRS constants.

Data Sources:

    - WAC 314-55-102 Quality assurance and quality control
    URL: <https://app.leg.wa.gov/wac/default.aspx?cite=314-55-102&pdf=true>

    - WAC 314-55-108 Pesticide action levels
    URL: <https://apps.leg.wa.gov/WAC/default.aspx?cite=314-55-108>

    - WAC 246-70-050 Quality assurance testing
    URL: <https://app.leg.wa.gov/WAC/default.aspx?cite=246-70-050&pdf=true>

    - WSLCB `TestName` List
    URL: <https://lcb.wa.gov/sites/default/files/publications/Cannabis/CCRS/Lab%20Test%20CSV%20TestName.pdf>

    - WSLCB Testing Requirements
    URL: <https://lcb.wa.gov/sites/default/files/publications/Cannabis/Education/Cannabis_Prod_Testing_Req_Past_vs_Current_infograph.pdf>
"""

# Datasets that comprise the CCRS.
CCRS_DATASETS = {
    'areas': {
        'dataset': 'Areas',
        'singular': 'area',
        'fields': {
            'AreaId': 'string',
            'LicenseeId': 'string',
            'Name': 'string',
            'IsQuarantine': 'boolean',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'contacts': {
        'dataset': 'Contacts',
        'singular': 'contact',
        'fields': {
            'ContactId': 'string',
            'LicenseeId': 'string',
            'IntegratorId': 'string',
            'FirstName': 'string',
            'MiddleInitial': 'string',
            'LastName': 'string',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
            'IsAdmin': 'boolean',
        },
        'date_fields': [
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'integrators': {
        'dataset': 'Integrator',
        'singular': 'integrator',
        'fields': {
            'IntegratorId': 'string',
            'Name': 'string',
            'IsActive': 'boolean',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'inventory': {
        'dataset': 'Inventory',
        'singular': 'inventory_item',
        'fields': {
            'LicenseeId': 'string',
            'InventoryId': 'string',
            'StrainId': 'string',
            'AreaId': 'string',
            'ProductId': 'string',
            'InventoryIdentifier': 'string',
            'InitialQuantity': 'float',
            'QuantityOnHand': 'float',
            'TotalCost': 'float',
            'IsMedical': 'boolean',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'updatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'updatedDate',
        ],
    },
    'inventory_adjustments': {
        'dataset': 'InventoryAdjustments',
        'singular': 'inventory_adjustment',
        'fields': {
            'InventoryAdjustmentId': 'string',
            'InventoryId': 'string',
            'InventoryAdjustmentReason': 'string',
            'AdjustmentDetail': 'string',
            'Quantity': 'float',
            'AdjustmentDate': 'datetime64',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'AdjustmentDate',
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'inventory_plant_transfers': {
        'dataset': 'InventoryPlantTransfers',
        'singular': 'inventory_adjustment',
        'fields': {
            'InventoryPlantTransferID': 'string',
            'InventoryTransferId': 'string',
            'PlantTransferId': 'string',
            'FromLicenseeId': 'string',
            'ToLicenseeId': 'string',
            'FromInventoryId': 'string',
            'ToInventoryId': 'string',
            'FromPlantId': 'string',
            'ToPlantId': 'string',
            'Quantity': 'float',
            'TransferDate': 'datetime64',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'TransferDate',
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'lab_results': {
        'dataset': 'LabResult',
        'singular': 'lab_result',
        'fields': {
            'LabResultId': 'string',
            'LabLicenseeId': 'string',
            'LicenseeId': 'string',
            'LabTestStatus': 'string',
            'InventoryId': 'string',
            'TestName': 'string',
            'TestDate': 'datetime64',
            'TestValue': 'float',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'string',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'TestDate',
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'licensees': {
        'dataset': 'Licensee',
        'singular': 'licensee',
        'fields': {
            'LicenseeId': 'string',
            'LicenseStatus': 'string',
            'UBI': 'string',
            'LicenseNumber': 'string',
            'Name': 'string',
            'DBA': 'string',
            'LicenseIssueDate': 'datetime64',
            'LicenseExpirationDate': 'datetime64',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'Address1': 'string',
            'Address2': 'string',
            'City': 'string',
            'State': 'string',
            'ZipCode': 'string',
            'County': 'string',
            'EmailAddress': 'string',
            'PhoneNumber': 'string',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'LicenseIssueDate',
            'LicenseExpirationDate',
            'UpdatedDate',
        ],
    },
    'plants': {
        'dataset': 'Plant',
        'singular': 'plant',
        'fields': {
            'PlantId': 'string',
            'LicenseeId': 'string',
            'AreaId': 'string',
            'StrainId': 'string',
            'PlantSource': 'string',
            'PlantState': 'string',
            'GrowthStage': 'string',
            'HarvestCycle': 'string',
            'MotherPlantId': 'string',
            'PlantIdentifier': 'string',
            'HarvestDate': 'datetime64',
            'IsMotherPlant': 'boolean',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'string',
        },
        'date_fields': [
            'CreatedDate',
            'HarvestDate',
            'UpdatedDate',
        ],
    },
    'plant_destructions': {
        'dataset': 'PlantDestructions',
        'singular': 'plant_destruction',
        'fields': {
            'PlantDestructionId': 'string',
            'PlantId': 'string',
            'DestructionReason': 'string',
            'DestructionMethod': 'string',
            'DestructionDate': 'datetime64',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'DestructionDate',
            'UpdatedDate',
        ],
    },
    'products': {
        'dataset': 'Product',
        'singular': 'product',
        'fields': {
            'ProductId': 'string',
            'LicenseeId': 'string',
            'InventoryType': 'string',
            'Name': 'string',
            'Description': 'string',
            'UnitWeightGrams': 'float',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'sale_headers': {
        'dataset': 'SaleHeader',
        'singular': 'sale',
        'fields': {
            'SaleHeaderId': 'string',
            'LicenseeId': 'string',
            'SoldToLicenseeId': 'string',
            'SaleType': 'string',
            'SaleDate': 'datetime64',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'SaleDate',
            'UpdatedDate',
        ],
    },
    'sale_details': {
        'dataset': 'SalesDetail',
        'singular': 'sale_item',
        'fields': {
            'SaleDetailId': 'string',
            'SaleHeaderId': 'string',
            'InventoryId': 'string',
            'PlantId': 'string',
            'Quantity': 'float',
            'UnitPrice': 'float',
            'Discount': 'float',
            'SalesTax': 'float',
            'OtherTax': 'float',
            'ExternalIdentifier': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    'strains': {
        'dataset': 'Strains',
        'singular': 'strain',
        'fields': {
            'StrainId': 'string',
            'StrainType': 'string',
            'Name': 'string',
            'IsDeleted': 'boolean',
            'CreatedBy': 'string',
            'CreatedDate': 'datetime64',
            'UpdatedBy': 'string',
            'UpdatedDate': 'datetime64',
        },
        'date_fields': [
            'CreatedDate',
            'UpdatedDate',
        ],
    },
    # Legacy: Transfers have been removed.
    'transfers': {
        'legacy': True,
        'dataset': 'Transfers',
        'singular': 'transfer',
        'fields': {
            'Serial': 'string',
            'SID': 'string',
            'Submitted Time': 'datetime64',
            'Completed Time': 'datetime64',
            'Modified Time': 'datetime64',
            'Draft': 'int',
            'UID': 'string',
            'Username': 'string',
            'Transportation Type': 'string',
            'Scheduled Transportation Date': 'datetime64',
            'UBI Number': 'string',
            'Origin License Number': 'string',
            'Origin Trade Name': 'string',
            'Origin License Address': 'string',
            'Origin License Phone': 'string',
            'Origin License E-mail Address': 'string',
            'Departure Date': 'datetime64',
            'Estimated Departure Time': 'time',
            'Arrival Date': 'datetime64',
            'Estimated Arrival Time': 'time',
            'Destination License Name': 'string',
            'Destination License Number': 'string',
            'Destination License Email': 'string',
            'Destination License Phone': 'string',
            'Destination License Address': 'string',
            'Items Shipped': 'string', # This is actually a complex CSV, TSV, etc.
        },
        'date_fields': [
            'Submitted Time',
            'Completed Time',
            'Modified Time',
            'Scheduled Transportation Date',
            'Departure Date',
            'Arrival Date',
        ],
    },
}

# CCRS analyses.
CCRS_ANALYSES = {
    'Foreign Matter': 'foreign_matter',
    'Heavy Metal': 'heavy_metals',
    'microbe': 'microbes',
    'Microbiological': 'microbes',
    'moisture': 'moisture',
    'Moisture': 'moisture',
    'Moisture Analysis': 'moisture',
    'Moistures Analysis': 'moisture',
    'Mycotoxin': 'mycotoxins',
    'mycotoxin': 'mycotoxins',
    'Pesticide': 'pesticides',
    'Potency': 'cannabinoids',
    'Residual Solvent': 'residual_solvents',
    'Terpene': 'terpenes',
    'terpene': 'terpenes',
    'Water Activity': 'water_activity',
    'water_activity': 'water_activity',
}

# CCRS analyte constants.
CCRS_ANALYTES = {
    'd9-THC': {'key': 'delta_9_thc', 'type': 'Potency', 'units': 'percent'},
    'CBD': {'key': 'cbd', 'type': 'Potency', 'units': 'percent'},
    'd9-THCA': {'key': 'thca', 'type': 'Potency', 'units': 'percent'},
    'Total d9-THC': {'key': 'total_thc', 'type': 'Potency', 'units': 'percent'},
    'CBDA': {'key': 'cbda', 'type': 'Potency', 'units': 'percent'},
    'Total CBD': {'key': 'total_cbd', 'type': 'Potency', 'units': 'percent'},
    'THC': {'key': 'delta_9_thc', 'type': 'Potency', 'units': 'percent'},
    'THCA': {'key': 'thca', 'type': 'Potency', 'units': 'percent'},
    'MAX THC': {'key': 'total_thc', 'type': 'Potency', 'units': 'percent'},
    'MAX CBD': {'key': 'total_cbd', 'type': 'Potency', 'units': 'percent'},
    'Microbial- I502 panel (3)-E.coli': {'key': 'e_coli', 'type': 'microbe', 'units': 'CFU/G'},
    'Microbial- I502 panel (3)-Bile Tolerant gram neg': {'key': 'btgn', 'type': 'microbe', 'units': 'CFU/G'},
    'Microbial- I502 panel (3)-Salmonella': {'key': 'salmonella', 'type': 'microbe', 'units': 'CFU/G'},
    'Moisture & Water Activity-Moisture(%)': {'key': 'moisture', 'type': 'moisture', 'units': 'percent'},
    'Moisture & Water Activity-Water Activity (Aw)(%)': {'key': 'water_activity', 'type': 'water_activity', 'units': 'aw'},
    'Mycotoxins-Total Aflatoxins': {'key': 'aflatoxins', 'type': 'mycotoxin', 'units': 'CFU/G'},
    'Mycotoxins-Ochratoxin A': {'key': 'ochratoxin', 'type': 'mycotoxin', 'units': 'CFU/G'},
    'Terpineol': {'key': 'terpineol', 'type': 'terpene', 'units': 'percent'},
    'Microbiological - BTGN (CFU/g)': {'key': 'btgn', 'type': 'Microbiological', 'units': 'CFU/g'},
    'Residual Solvent - Butanes (ppm)': {'key': 'butanes', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Propanes (ppm)': {'key': 'propanes', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Benzene (ppm)': {'key': 'benzene', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Potency - D9THCA (%)': {'key': 'thca', 'type': 'Potency', 'units': 'percent'},
    'Potency - Total THC (%)': {'key': 'total_thc', 'type': 'Potency', 'units': 'percent'},
    'Potency - CBDA (%)': {'key': 'cbda', 'type': 'Potency', 'units': 'percent'},
    'Foreign Matter - Seeds or Other (%)': {'key': 'seeds_or_other', 'type': 'Foreign Matter', 'units': 'percent'},
    'Potency - Total CBD (%)': {'key': 'total_cbd', 'type': 'Potency', 'units': 'percent'},
    'Potency - D9THC (%)': {'key': '9_thc', 'type': 'Potency', 'units': 'percent'},
    'Potency - CBD (%)': {'key': 'cbd', 'type': 'Potency', 'units': 'percent'},
    'Foreign Matter - Stems (%)': {'key': 'stems', 'type': 'Foreign Matter', 'units': 'percent'},
    'Residual Solvent - Toluene (ppm)': {'key': 'toluene', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Heptanes (ppm)': {'key': 'heptanes', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Microbiological - Salmonella (CFU/g)': {'key': 'salmonella', 'type': 'Microbiological', 'units': 'CFU/g'},
    'Microbiological - Ste (CFU/g)': {'key': 'stec', 'type': 'Microbiological', 'units': 'CFU/g'},
    'Moisture Analysis - Water Activity (aw)': {'key': 'water_activity', 'type': 'Moisture Analysis', 'units': 'aw'},
    'Mycotoxin - Ochratoxin A (ppb)': {'key': 'ochratoxin', 'type': 'Mycotoxin', 'units': 'ppb'},
    'Moistures Analysis - Moisture Content (%)': {'key': 'moisture_content', 'type': 'Moistures Analysis', 'units': 'percent'},
    'Mycotoxin - Total Aflatoxins (ppb)': {'key': 'total_aflatoxins', 'type': 'Mycotoxin', 'units': 'ppb'},
    'Residual Solvent - Hexanes (ppm)': {'key': 'hexanes', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Acetone (ppm)': {'key': 'acetone', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Xylenes (ppm)': {'key': 'xylenes', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Methanol (ppm)': {'key': 'methanol', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Isopropanol (ppm)': {'key': 'isopropanol', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Chloroform (ppm)': {'key': 'chloroform', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Dichloromethane (ppm)': {'key': 'dichloromethane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Ethyl Acetate (ppm)': {'key': 'ethyl_acetate', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Cyclohexane (ppm)': {'key': 'cyclohexane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Potency - CBDV (%)': {'key': 'cbdv', 'type': 'Potency', 'units': 'percent'},
    'Potency - THCV (%)': {'key': 'thcv', 'type': 'Potency', 'units': 'percent'},
    'Potency - CBG (%)': {'key': 'cbg', 'type': 'Potency', 'units': 'percent'},
    'Potency - CBGA (%)': {'key': 'cbga', 'type': 'Potency', 'units': 'percent'},
    'Potency - CBN (%)': {'key': 'cbn', 'type': 'Potency', 'units': 'percent'},
    'Potency - D8THC (%)': {'key': 'delta_8_thc', 'type': 'Potency', 'units': 'percent'},
    'Potency - CBC (%)': {'key': 'cbc', 'type': 'Potency', 'units': 'percent'},
    'Potency - THCA (%)': {'key': 'thca', 'type': 'Potency', 'units': 'percent'},
    'Mycotoxin - Total Aflatoxins (ug/kg)': {'key': 'total_aflatoxins', 'type': 'Mycotoxin', 'units': 'ug/kg'},
    'Mycotoxin - Ochratoxin A (ug/kg)': {'key': 'ochratoxin', 'type': 'Mycotoxin', 'units': 'ug/kg'},
    'Residual Solvent - Acetone (ug/g)': {'key': 'acetone', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Acetonitrile (ug/g)': {'key': 'acetonitrile', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Benzene (ug/g)': {'key': 'benzene', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Butanes (ug/g)': {'key': 'butane', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Chloroform (ug/g)': {'key': 'chloroform', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Cyclohexane (ug/g)': {'key': 'cyclohexane', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Dichloromethane (ug/g)': {'key': 'dichloromethane', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Ethanol (ug/g)': {'key': 'ethanol', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Ethanol (ppm)': {'key': 'ethanol', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Ethanol': {'key': 'ethanol', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Total Butanes': {'key': 'butane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Ethyl Acetate (ug/g)': {'key': 'ethyl_acetate', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Ethyl Ether (ug/g)': {'key': 'ethyl_ether', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Heptanea (ug/g)': {'key': 'heptane', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Hexanes (ug/g)': {'key': 'hexanes', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Isobutane (ug/g)': {'key': 'isobutane', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Methanol (ug/g)': {'key': 'methanol', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Pentanes (ug/g)': {'key': 'pentanes', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - 2-Propanol (IPA) (ug/g)': {'key': '2_propanol', 'type': 'Residual Solvent', 'units': 'IPA '},
    'Residual Solvent - Propanes (ug/g)': {'key': 'propanes', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Toluene (ug/g)': {'key': 'toluene', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Xylenes (ug/g)': {'key': 'xylenes', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Moisture Analysis - Moisture Content (%)': {'key': 'moisture_content', 'type': 'Moisture Analysis', 'units': 'percent'},
    'Residual Solvent - Heptanes (ug/g)': {'key': 'heptanes', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Residual Solvent - Isopropanol (ug/g)': {'key': 'isopropanol', 'type': 'Residual Solvent', 'units': 'ug/g'},
    'Potency - CBD (mg/g)': {'key': 'cbd', 'type': 'Potency', 'units': 'mg/g'},
    'Potency - CBDA (mg/g)': {'key': 'cbda', 'type': 'Potency', 'units': 'mg/g'},
    'Potency - D9THC (mg/g)': {'key': 'delta_9_thc', 'type': 'Potency', 'units': 'mg/g'},
    'Potency - D9THCA (mg/g)': {'key': 'thca', 'type': 'Potency', 'units': 'mg/g'},
    'Potency - Total THC (mg/g)': {'key': 'total_thc', 'type': 'Potency', 'units': 'mg/g'},
    'Potency - Total CBD (mg/g)': {'key': 'total_cbd', 'type': 'Potency', 'units': 'mg/g'},
    'Potency - CBD (mg/mL)': {'key': 'cbd', 'type': 'Potency', 'units': 'mg/mL'},
    'Potency - CBDA (mg/mL)': {'key': 'cbda', 'type': 'Potency', 'units': 'mg/mL'},
    'Potency - D9THC (mg/mL)': {'key': 'delta_9_thc', 'type': 'Potency', 'units': 'mg/mL'},
    'Potency - D9THCA (mg/mL)': {'key': 'thca', 'type': 'Potency', 'units': 'mg/mL'},
    'Potency - Total THC (mg/mL)': {'key': 'total_thc', 'type': 'Potency', 'units': 'mg/mL'},
    'Potency - Total CBD (mg/mL)': {'key': 'total_cbd', 'type': 'Potency', 'units': 'mg/mL'},
    'Microbiological - BTGN (CFU/G)': {'key': 'btgn', 'type': 'Microbiological', 'units': 'CFU/G'},
    'Mycotoxin - Ochratoxin (CFU/G)': {'key': 'ochratoxin', 'type': 'Mycotoxin', 'units': 'CFU/G'},
    'Microbiological - STEC (CFU/G)': {'key': 'stec', 'type': 'Microbiological', 'units': 'CFU/G'},
    'Microbiological - STEC (CFU/g)': {'key': 'stec', 'type': 'Microbiological', 'units': 'CFU/G'},
    'Microbiological - Salmonella (CFU/G)': {'key': 'salmonella', 'type': 'Microbiological', 'units': 'CFU/G'},
    'Mycotoxin - Aflatoxins (CFU/G)': {'key': 'aflatoxins', 'type': 'Mycotoxin', 'units': 'CFU/G'},
    'Residual Solvent - Benzene (PPM)': {'key': 'benzene', 'type': 'Residual Solvent', 'units': 'PPM'},
    'Residual Solvent - Dichloro-Methane (ppm)': {'key': 'dichloromethane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Heptane (ppm)': {'key': 'heptane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - n-Hexane (ppm)': {'key': 'n_hexane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Pentanes (ppm)': {'key': 'pentanes', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Residual Solvent - Propane (ppm)': {'key': 'propane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Potency - D9THC (mg/serving)': {'key': 'delta_9_thc', 'type': 'Potency', 'units': 'mg/serving'},
    'Potency - D9THCA (mg/serving)': {'key': 'thca', 'type': 'Potency', 'units': 'mg/serving'},
    'Potency - CBD (mg/serving)': {'key': 'cbd', 'type': 'Potency', 'units': 'mg/serving'},
    'Potency - CBDA (mg/serving)': {'key': 'cbda', 'type': 'Potency', 'units': 'mg/serving'},
    'Potency - Total THC (mg/serving)': {'key': 'total_thc', 'type': 'Potency', 'units': 'mg/serving'},
    'Potency - Total CBD (mg/serving)': {'key': 'total_cbd', 'type': 'Potency', 'units': 'mg/serving'},
    'Potency - D9THC ()': {'key': 'delta_9_thc', 'type': 'Potency', 'units': ''},
    'Potency - D9THCA ()': {'key': 'thca', 'type': 'Potency', 'units': ''},
    'Potency - CBD ()': {'key': 'cbd', 'type': 'Potency', 'units': ''},
    'Potency - CBDA ()': {'key': 'cbda', 'type': 'Potency', 'units': ''},
    'Potency - Total THC ()': {'key': 'total_thc', 'type': 'Potency', 'units': ''},
    'Potency - Total CBD ()': {'key': 'total_cbd', 'type': 'Potency', 'units': ''},
    'Potency - THCA (mg/serving)': {'key': 'thca', 'type': 'Potency', 'units': 'mg/serving'},
    'Heavy Metal - Arsenic (ppm)': {'key': 'arsenic', 'type': 'Heavy Metal', 'units': 'ppm'},
    'Heavy Metal - Cadmium (ppm)': {'key': 'cadmium', 'type': 'Heavy Metal', 'units': 'ppm'},
    'Heavy Metal - Lead (ppm)': {'key': 'lead', 'type': 'Heavy Metal', 'units': 'ppm'},
    'Heavy Metal - Mercury (ppm)': {'key': 'mercury', 'type': 'Heavy Metal', 'units': 'ppm'},
    'Heavy Metal - Arsenic (ug/g)': {'key': 'arsenic', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Heavy Metal - Mercury (ug/g)': {'key': 'mercury', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Heavy Metal - Cadmium (ug/g)': {'key': 'cadmium', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Heavy Metal - Lead (ug/g)': {'key': 'lead', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Arsenic': {'key': 'arsenic', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Cadmium': {'key': 'cadmium', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Lead': {'key': 'lead', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Mercury': {'key': 'mercury', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Pb': {'key': 'lead', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'As': {'key': 'arsenic', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Hg': {'key': 'mercury', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Cd': {'key': 'cadmium', 'type': 'Heavy Metal', 'units': 'ug/g'},
    'Pesticide - Abamectin (ppm)': {'key': 'abamectin', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Acephate (ppm)': {'key': 'acephate', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Acequinocyl (ppm)': {'key': 'acequinocyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Acetamiprid (ppm)': {'key': 'acetamiprid', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Aldicarb (ppm)': {'key': 'aldicarb', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Azoxystrobin (ppm)': {'key': 'azoxystrobin', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Bifenazate (ppm)': {'key': 'bifenazate', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Bifenthrin (ppm)': {'key': 'bifenthrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Boscalid (ppm)': {'key': 'boscalid', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Carbaryl (ppm)': {'key': 'carbaryl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Carbofuran (ppm)': {'key': 'carbofuran', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Chlorantraniliprole (ppm)': {'key': 'chlorantraniliprole', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Chlorfenapyr (ppm)': {'key': 'chlorfenapyr', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Chlorpyrifos (ppm)': {'key': 'chlorpyrifos', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Clofentizine (ppm)': {'key': 'clofentizine', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Cyfluthrin (ppm)': {'key': 'cyfluthrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Cypermethrin (ppm)': {'key': 'cypermethrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Daminozide (ppm)': {'key': 'daminozide', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - DDVP (dichlorvos) (ppm)': {'key': 'ddvp', 'type': 'Pesticide', 'units': 'dichlorvos '},
    'Pesticide - Diazinon (ppm)': {'key': 'diazinon', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Dimethoate (ppm)': {'key': 'dimethoate', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Ethoprophos (ppm)': {'key': 'ethoprophos', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Etofenprox (ppm)': {'key': 'etofenprox', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Etoxazole (ppm)': {'key': 'etoxazole', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Fenoxycarb (ppm)': {'key': 'fenoxycarb', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Fenpyroximate (ppm)': {'key': 'fenpyroximate', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Fipronil (ppm)': {'key': 'fipronil', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Flonicamid (ppm)': {'key': 'flonicamid', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Fludioxonil (ppm)': {'key': 'fludioxonil', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Hexythiazox (ppm)': {'key': 'hexythiazox', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Imazalil (ppm)': {'key': 'imazalil', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Imidacloprid (ppm)': {'key': 'imidacloprid', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Kresoxim-Methyl (ppm)': {'key': 'kresoxim_methyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Malathion (ppm)': {'key': 'malathion', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Metalaxyl (ppm)': {'key': 'metalaxyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Methiocarb (ppm)': {'key': 'methiocarb', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Methomyl (ppm)': {'key': 'methomyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Methyl parathion (ppm)': {'key': 'methyl_parathion', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - MGK-264 (ppm)': {'key': 'mgk_264', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Myclobutanil (ppm)': {'key': 'myclobutanil', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Naled (ppm)': {'key': 'naled', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Oxamyl (ppm)': {'key': 'oxamyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Paclobutrazol (ppm)': {'key': 'paclobutrazol', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Permethrins (ppm)': {'key': 'permethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Phosmet (ppm)': {'key': 'phosmet', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Piperonyl butoxide (ppm)': {'key': 'piperonyl_butoxide', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Prallethrin (ppm)': {'key': 'prallethrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Propiconazole (ppm)': {'key': 'propiconazole', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Propoxur (ppm)': {'key': 'propoxur', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Pyrethrins (ppm)': {'key': 'pyrethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Pyridaben (ppm)': {'key': 'pyridaben', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Spinosad (ppm)': {'key': 'spinosad', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Spiromesifen (ppm)': {'key': 'spiromesifen', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Spirotetramat (ppm)': {'key': 'spirotetramat', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Spiroxamine (ppm)': {'key': 'spiroxamine', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Tebuconazole (ppm)': {'key': 'tebuconazole', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Thiacloprid (ppm)': {'key': 'thiacloprid', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Thiamethoxam (ppm)': {'key': 'thiamethoxam', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Trifloxystrobin (ppm)': {'key': 'trifloxystrobin', 'type': 'Pesticide', 'units': 'ppm'},
    'Etofenprox': {'key': 'etofenprox', 'type': 'Pesticide', 'units': 'ppm'},
    'Bifenthrin': {'key': 'bifenthrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Moisture Analysis - Water Activity (Aw)': {'key': 'water_activity', 'type': 'Moisture', 'units': 'Aw'},
    'Moisture Content': {'key': 'moisture_content', 'type': 'Moisture Analysis', 'units': 'percent'},
    'Water Activity': {'key': 'water_activity', 'type': 'Moisture', 'units': 'Aw'},
    'Foreign Matter - IHE (Boolean)': {'key': 'soil', 'type': 'Foreign Matter', 'units': 'boolean'},
    'Foreign Matter - IHE (ea.)': {'key': 'soil', 'type': 'Foreign Matter', 'units': 'ea'},
    'Foreign Matter - HIE (ea)': {'key': 'foreign_matter', 'type': 'Foreign Matter', 'units': 'ea'},
    'Foreign Matter - IHE (%)': {'key': 'foreign_matter', 'type': 'Foreign Matter', 'units': 'percent'},
    'Foreign I.H.E.': {'key': 'foreign_matter', 'type': 'Foreign Matter', 'units': 'percent'},
    'Foreign Matter': {'key': 'foreign_matter', 'type': 'Foreign Matter', 'units': 'percent'},
    'Foreign Matter (%)': {'key': 'foreign_matter', 'type': 'Foreign Matter', 'units': 'percent'},
    'Stems (%)': {'key': 'stems', 'type': 'Foreign Matter', 'units': 'percent'},
    'Stems': {'key': 'stems', 'type': 'Foreign Matter', 'units': 'percent'},
    'Foreign matter': {'key': 'foreign_matter', 'type': 'Foreign Matter', 'units': 'percent'},
    'Mycotoxin - Aflatoxins (ug/kg)': {'key': 'total_aflatoxins', 'type': 'Mycotoxin', 'units': 'ug/kg'},
    'Mycotoxin - Ochratoxin (ug/kg)': {'key': 'ochratoxin_a', 'type': 'Mycotoxin', 'units': 'ug/kg'},
    'Mycotoxin - Aflatoxins (ï¿½g/kg)': {'key': 'total_aflatoxins', 'type': 'Mycotoxin', 'units': 'ug/kg'},
    'Mycotoxin - Ochratoxin (ï¿½g/kg)': {'key': 'ochratoxin_a', 'type': 'Mycotoxin', 'units': 'ug/kg'},
    'Aflatoxins': {'key': 'ochratoxin', 'type': 'Mycotoxin', 'units': 'CFU/G'},
    'Total Aflatoxins': {'key': 'total_aflatoxins', 'type': 'Mycotoxin', 'units': 'CFU/G'},
    'Ochratoxin A': {'key': 'ochratoxin_a', 'type': 'Mycotoxin', 'units': 'CFU/G'},
    'BTGN': {'key': 'btgn', 'type': 'Microbiological', 'units': 'CFU/G'},
    'Salmonella': {'key': 'salmonella', 'type': 'Microbiological', 'units': 'CFU/G'},
    'E Coli': {'key': 'e_coli', 'type': 'Microbiological', 'units': 'CFU/G'},
    'STEC': {'key': 'stec', 'type': 'Microbiological', 'units': 'CFU/G'},
    'Salmonella spp.': {'key': 'salmonella', 'type': 'Microbiological', 'units': 'CFU/G'},
    'Acetone': {'key': 'acetone', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Toluene': {'key': 'toluene', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Total Hexanes': {'key': 'hexane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Total Xylenes': {'key': 'total_xylenes', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Total Propanes': {'key': 'propane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Benzene': {'key': 'benzene', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Total Heptanes': {'key': 'heptane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'iso-Propanol': {'key': 'isopropanol', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Chloroform': {'key': 'chloroform', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Cyclohexane': {'key': 'cyclohexane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Dichloromethane': {'key': 'dichloromethane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Total Pentanes': {'key': 'pentane', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Ethyl_Acetate': {'key': 'ethyl_acetate', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Methanol': {'key': 'methanol', 'type': 'Residual Solvent', 'units': 'ppm'},
    'Piperonyl Butoxide': {'key': 'piperonyl_butoxide', 'type': 'Pesticide', 'units': 'ppm'},
    'Cyfluthrin': {'key': 'cyfluthrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Chlorfenapyr': {'key': 'chlorfenapyr', 'type': 'Pesticide', 'units': 'ppm'},
    'Cypermethrin': {'key': 'cypermethrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Total MGK 264': {'key': 'total_mgk_264', 'type': 'Pesticide', 'units': 'ppm'},
    'Total Permethrins': {'key': 'total_permethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Carbofuran': {'key': 'carbofuran', 'type': 'Pesticide', 'units': 'ppm'},
    'Total Pyrethrins': {'key': 'pyrethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Azoxystrobin': {'key': 'azoxystrobin', 'type': 'Pesticide', 'units': 'ppm'},
    'Fludioxonil': {'key': 'fludioxonil', 'type': 'Pesticide', 'units': 'ppm'},
    'Propiconazole': {'key': 'propiconazole', 'type': 'Pesticide', 'units': 'ppm'},
    'Ethoprophos': {'key': 'ethoprophos', 'type': 'Pesticide', 'units': 'ppm'},
    'Imidacloprid': {'key': 'imidacloprid', 'type': 'Pesticide', 'units': 'ppm'},
    'Propoxur': {'key': 'propoxur', 'type': 'Pesticide', 'units': 'ppm'},
    'Myclobutanil': {'key': 'myclobutanil', 'type': 'Pesticide', 'units': 'ppm'},
    'Hexythiazox': {'key': 'hexythiazox', 'type': 'Pesticide', 'units': 'ppm'},
    'Paclobutrazol': {'key': 'paclobutrazol', 'type': 'Pesticide', 'units': 'ppm'},
    'Total Abamectin': {'key': 'total_abamectin', 'type': 'Pesticide', 'units': 'ppm'},
    'Dichlorvos': {'key': 'dichlorvos', 'type': 'Pesticide', 'units': 'ppm'},
    'Trifloxystrobin': {'key': 'trifloxystrobin', 'type': 'Pesticide', 'units': 'ppm'},
    'Tebuconazole': {'key': 'tebuconazole', 'type': 'Pesticide', 'units': 'ppm'},
    'Thiamethoxam': {'key': 'thiamethoxam', 'type': 'Pesticide', 'units': 'ppm'},
    'Acequinocyl': {'key': 'acequinocyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Diazinon': {'key': 'diazinon', 'type': 'Pesticide', 'units': 'ppm'},
    'Fenoxycarb': {'key': 'fenoxycarb', 'type': 'Pesticide', 'units': 'ppm'},
    'Methiocarb': {'key': 'methiocarb', 'type': 'Pesticide', 'units': 'ppm'},
    'Etoxazole': {'key': 'etoxazole', 'type': 'Pesticide', 'units': 'ppm'},
    'Carbaryl': {'key': 'carbaryl', 'type': 'Pesticide', 'units': 'ppm'},
    'Fenpyroximate': {'key': 'fenpyroximate', 'type': 'Pesticide', 'units': 'ppm'},
    'Spirotetramat': {'key': 'spirotetramat', 'type': 'Pesticide', 'units': 'ppm'},
    'Methomyl': {'key': 'methomyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Clofentezine': {'key': 'clofentezine', 'type': 'Pesticide', 'units': 'ppm'},
    'Aldicarb': {'key': 'aldicarb', 'type': 'Pesticide', 'units': 'ppm'},
    'Spiromesifen': {'key': 'spiromesifen', 'type': 'Pesticide', 'units': 'ppm'},
    'Boscalid': {'key': 'boscalid', 'type': 'Pesticide', 'units': 'ppm'},
    'Imazalil': {'key': 'imazalil', 'type': 'Pesticide', 'units': 'ppm'},
    'Acephate': {'key': 'acephate', 'type': 'Pesticide', 'units': 'ppm'},
    'Fipronil': {'key': 'fipronil', 'type': 'Pesticide', 'units': 'ppm'},
    'Total Spinosads': {'key': 'spinosad', 'type': 'Pesticide', 'units': 'ppm'},
    'Daminozide': {'key': 'daminozide', 'type': 'Pesticide', 'units': 'ppm'},
    'Metalaxyl': {'key': 'metalaxyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Bifenazate': {'key': 'bifenazate', 'type': 'Pesticide', 'units': 'ppm'},
    'Spiroxamine': {'key': 'spiroxamine', 'type': 'Pesticide', 'units': 'ppm'},
    'Naled': {'key': 'naled', 'type': 'Pesticide', 'units': 'ppm'},
    'Dimethoate': {'key': 'dimethoate', 'type': 'Pesticide', 'units': 'ppm'},
    'Kresoxim-methyl': {'key': 'kresoxim_methyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Oxamyl': {'key': 'oxamyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pyridaben': {'key': 'pyridaben', 'type': 'Pesticide', 'units': 'ppm'},
    'Malathion': {'key': 'malathion', 'type': 'Pesticide', 'units': 'ppm'},
    'Methyl parathion': {'key': 'methyl_parathion', 'type': 'Pesticide', 'units': 'ppm'},
    'Flonicamid': {'key': 'flonicamid', 'type': 'Pesticide', 'units': 'ppm'},
    'Acetamiprid': {'key': 'acetamiprid', 'type': 'Pesticide', 'units': 'ppm'},
    'Prallethrin': {'key': 'prallethrin', 'type': 'Pesticide', 'units': 'ppm'},
    'Phosmet': {'key': 'phosmet', 'type': 'Pesticide', 'units': 'ppm'},
    'Chlorpyrifos': {'key': 'chlorpyrifos', 'type': 'Pesticide', 'units': 'ppm'},
    'Chlorantraniliprole': {'key': 'chlorantraniliprole', 'type': 'Pesticide', 'units': 'ppm'},
    'Thiacloprid': {'key': 'thiacloprid', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Flonicamid (ug/g)': {'key': 'flonicamid', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Fludioxonil (ug/g)': {'key': 'fludioxonil', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Hexythiazox (ug/g)': {'key': 'hexythiazox', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Imazalil (ug/g)': {'key': 'imazalil', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Imidacloprid (ug/g)': {'key': 'imidacloprid', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Kresoxim-Methyl (ug/g)': {'key': 'kresoxim_methyl', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Malathion (ug/g)': {'key': 'malathion', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Metalaxyl (ug/g)': {'key': 'metalaxyl', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Methiocarb (ug/g)': {'key': 'methiocarb', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Methomyl (ug/g)': {'key': 'methomyl', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Methyl parathion (ug/g)': {'key': 'methyl_parathion', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - MGK-264 (ug/g)': {'key': 'mgk_264', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Myclobutanil (ug/g)': {'key': 'myclobutanil', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Naled (ug/g)': {'key': 'naled', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Oxamyl (ug/g)': {'key': 'oxamyl', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Paclobutrazol (ug/g)': {'key': 'paclobutrazol', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Permethrins (ug/g)': {'key': 'permethrins', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Phosmet (ug/g)': {'key': 'phosmet', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Piperonyl butoxide (ug/g)': {'key': 'piperonyl_butoxide', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Prallethrin (ug/g)': {'key': 'prallethrin', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Propiconazole (ug/g)': {'key': 'propiconazole', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Propoxur (ug/g)': {'key': 'propoxur', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Pyrethrins (ug/g)': {'key': 'pyrethrins', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Pyridaben (ug/g)': {'key': 'pyridaben', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Spinosad (ug/g)': {'key': 'spinosad', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Spiromesifen (ug/g)': {'key': 'spiromesifen', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Spirotetramat (ug/g)': {'key': 'spirotetramat', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Spiroxamine (ug/g)': {'key': 'spiroxamine', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Tebuconazole (ug/g)': {'key': 'tebuconazole', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Thiacloprid (ug/g)': {'key': 'thiacloprid', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Thiamethoxam (ug/g)': {'key': 'thiamethoxam', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Trifloxystrobin (ug/g)': {'key': 'trifloxystrobin', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Acephate (ug/g)': {'key': 'acephate', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Acequinocyl (ug/g)': {'key': 'acequinocyl', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Acetamiprid (ug/g)': {'key': 'acetamiprid', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Aldicarb (ug/g)': {'key': 'aldicarb', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Bifenazate (ug/g)': {'key': 'bifenazate', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Bifenthrin (ug/g)': {'key': 'bifenthrin', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Boscalid (ug/g)': {'key': 'boscalid', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Carbaryl (ug/g)': {'key': 'carbaryl', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Carbofuran (ug/g)': {'key': 'carbofuran', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Chlorfenapyr (ug/g)': {'key': 'chlorfenapyr', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Chlorpyrifos (ug/g)': {'key': 'chlorpyrifos', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Clofentizine (ug/g)': {'key': 'clofentizine', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Cyfluthrin (ug/g)': {'key': 'cyfluthrin', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Cypermethrin (ug/g)': {'key': 'cypermethrin', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Daminozide (ug/g)': {'key': 'daminozide', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - DDVP (dichlorvos) (ug/g)': {'key': 'ddvp_dichlorvos', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Diazinon (ug/g)': {'key': 'diazinon', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Dimethoate (ug/g)': {'key': 'dimethoate', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Ethoprophos (ug/g)': {'key': 'ethoprophos', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Etofenprox (ug/g)': {'key': 'etofenprox', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Etoxazole (ug/g)': {'key': 'etoxazole', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Fenoxycarb (ug/g)': {'key': 'fenoxycarb', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Fenpyroximate (ug/g)': {'key': 'fenpyroximate', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Fipronil (ug/g)': {'key': 'fipronil', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Abamectin (ug/g)': {'key': 'abamectin', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Azoxystrobin (ug/g)': {'key': 'azoxystrobin', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Chlorantraniliprole (ug/g)': {'key': 'chlorantraniliprole', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - DDVP (dichlorvos)': {'key': 'ddvp_dichlorvos', 'type': 'Pesticide', 'units': 'ug/g'},
    'Pesticide - Clofentezine (ppm)': {'key': 'clofentezine', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Dichlorvos (DDVP) (ppm)': {'key': 'dichlorvos_ddvp', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Kresoxin-methyl (ppm)': {'key': 'kresoxin_methyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Permethrins (total) (ppm)': {'key': 'permethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Phosemet (ppm)': {'key': 'phosemet', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Piperonyl Butoxide (ppm)': {'key': 'piperonyl_butoxide', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Pyrethrin I (ppm)': {'key': 'pyrethrin_i', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Spinosads (total) (ppm)': {'key': 'spinosad', 'type': 'Pesticide', 'units': 'ppm'},
    'Pesticide - Piperonyl Butoxide (ug/g)': {'key': 'piperonyl_butoxide', 'type': 'Pesticide', 'units': 'ug/g'},
    'Loss on Drying': {'key': 'moisture_content', 'type': 'Moistures Analysis', 'units': 'percent'},
    'Abamectin B1a': {'key': 'abamectin', 'type': 'Pesticide', 'units': 'ppm'},
    'Pyrethrins I': {'key': 'pyrethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Pyrethrins (Total)': {'key': 'pyrethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'DDVP (dichlorvos)': {'key': 'ddvp', 'type': 'Pesticide', 'units': 'dichlorvos '},
    'Kresoxim-Methyl': {'key': 'kresoxim_methyl', 'type': 'Pesticide', 'units': 'ppm'},
    'Clofentizine': {'key': 'clofentizine', 'type': 'Pesticide', 'units': 'ppm'},
    'Piperonyl butoxide': {'key': 'piperonyl_butoxide', 'type': 'Pesticide', 'units': 'ppm'},
    'Permethrins': {'key': 'permethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Spinosad': {'key': 'spinosad', 'type': 'Pesticide', 'units': 'ppm'},
    'MGK-264': {'key': 'mgk_264', 'type': 'Pesticide', 'units': 'ppm'},
    'Pyrethrins': {'key': 'pyrethrins', 'type': 'Pesticide', 'units': 'ppm'},
    'Abamectin': {'key': 'abamectin', 'type': 'Pesticide', 'units': 'ppm'},
}

# CCRS plant constants.
CCRS_PLANT_GROWTH_STAGES = ['Vegetative', 'Immature', 'Flowering']
CCRS_PLANT_HARVEST_CYCLES = [3,  6,  9, 12]
CCRS_PLANTS_SOURCES = ['Clone', 'Seed']
CCRS_PLANT_STATES = [
    'Growing',
    'Inventory',
    'Drying',
    'Harvested',
    'Destroyed',
    'Sold',
]



"""
Cannabis License Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Cannabis license data model.
"""
# Standard imports.
from dataclasses import dataclass

# Internal imports.
from .base import Model


@dataclass
class CannabisLicense(Model):
    """A state-issued cannabis license."""
    _collection = 'organizations/%s/licenses'
    active_date: str = ''
    expiration_date: str = ''
    license_number: str = ''
    license_type: str = ''



"""
Metrc Exceptions | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/8/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Exceptions used when interfacing with the Metrc API.
"""


class MetrcAPIError(Exception):
    """A primary error raised by the Metrc API.
    Insufficient permissions for a request typically
    result in a 401 unauthorized error.
    """

    def __init__(self, response):
        message = self.get_response_messages(response)
        super().__init__(message)
        self.response = response

    def get_response_messages(self, response):
        """Extract error messages from a Metrc API response.
        Args:
            response (Response): A request response from the Metrc API.
        Returns:
            (str): Returns any error messages.
        """
        try:
            errors = response.json()
            if isinstance(errors, list):
                try:
                    message = '\n'.join(errors)
                except TypeError:
                    message = '\n'.join([x.get('message') for x in errors])
            elif isinstance(errors, dict):
                message = errors.get('Message', errors.get('message'))
            else:
                message = response.text
        except (AttributeError, KeyError, ValueError):
            message = 'Unknown Metrc API error'
        return message



"""
Authentication Logic | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 1/22/2021
Updated: 7/2/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Authentication mechanisms for the Cannlytics API,
specifically API request authentication and verification.
"""
# Standard imports.
import hmac
from hashlib import sha256

# Internal imports.
from ..firebase import (
    get_custom_claims,
    get_document,
    verify_session_cookie,
    verify_token,
)


def authenticate_request(request):
    """Verifies that the user has authenticated with a Firebase ID token
    or passed a valid API key in an `Authentication: Bearer <token>` header.
    Args:
        request: An instance of `django.http.HttpRequest` or
            `rest_framework.request.Request`.
    Returns:
        claims (dict): A dictionary of the user's custom claims, including
            the user's `uid`.
    """
    claims = {}
    try:
        session_cookie = request.COOKIES.get('__session')
        if session_cookie is None:
            session_cookie = request.session.get('__session')
        claims = verify_session_cookie(session_cookie, check_revoked=True)
    except:
        try:
            authorization = request.META['HTTP_AUTHORIZATION']
            key = authorization.split(' ').pop()
            try:
                claims = get_user_from_api_key(key)
            except:
                claims = verify_token(key)
        except:
            claims = {}
    return claims


def get_user_from_api_key(api_key: str) -> dict:
    """Identify a user given an API key.
    Args:
        api_key (str): An API key to identify a given user.
    Returns:
        (dict): Any user data found, with an empty dictionary if there
            is no user found.
    """
    app_secret = get_document('admin/api')['app_secret_key']
    code = sha256_hmac(app_secret, api_key)
    key_data = get_document(f'admin/api/api_key_hmacs/{code}')
    uid = key_data['uid']
    user_claims = get_custom_claims(uid)
    user_claims['permissions'] = key_data['permissions']
    user_claims['uid'] = uid
    return user_claims


def sha256_hmac(secret, message):
    """Create a SHA256-HMAC (hash-based message authentication code).
    Args:
        secret (str): A server-side app secret.
        message (str): The client's secret.
    Returns:
        (str): An HMAC string.
    Credit: https://stackoverflow.com/a/66958131/5021266
    """
    return hmac.new(bytes(secret, 'UTF-8'), message.encode(), sha256).hexdigest()



"""
Constants | Cannlytics
Copyright (c) 2021-2022 Cannlytics and the Cannabis Data Science Team

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/8/2021
Updated: 11/28/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: This module contains useful constants. In particular, there
are maps of encountered `ANALYSES`, `ANALYTES`, `STANDARD_FIELDS`,
`STANDARD_UNITS, `PRODUCT_TYPES`, and `STRAINS` to standard keys.
There are also `CODINGS` to help normalize result values.
The remaining useful constants are: `DECARB`, `DEFAULT_HEADERS`,
`RANDOM_STRING_CHARS`, `states`, `state_names`, `state_time_zones`.
"""

# Standard analysis key map.
# Optional: Load known analyses and analytes from
# the Cannlytics Database via the Cannlytics API.
STANDARD_ANALYSES = {
    'cannabinoids': {'name': 'Cannabinoids'},
    'terpenes': {'name': 'Terpenes'},
    'residual_solvents': {'name': 'Residual Solvents'},
    'pesticides': {'name': 'Pesticides'},
    'microbes': {'name': 'Microbes'},
    'mycotoxins': {'name': 'Mycotoxins'},
    'heavy_metals': {'name': 'Heavy Metals'},
    'foreign_matter': {'name': 'Foreign Matter'},
    'moisture_content': {'name': 'Moisture Content'},
    'water_activity': {'name': 'Water Activity'},
}

# A map of encountered analyses to their standardized analysis.
ANALYSES = {
    'cannabinoids': 'cannabinoids',
    'cannabinoid': 'cannabinoids',
    'Cannabinoid': 'cannabinoids',
    'POT': 'cannabinoids',
    'Potency': 'cannabinoids',
    'Potency Analysis by HPLC': 'cannabinoids',
    'Potency Test Result': 'cannabinoids',
    'CANNABINOID': 'cannabinoids',
    'foreign_materials': 'foreign_matter',
    'foreign_matter': 'foreign_matter',
    'Foreign Material': 'foreign_matter',
    'foreign_material': 'foreign_matter',
    'Filth and Foreign Material': 'foreign_matter',
    'Filth & Foreign Material': 'foreign_matter',
    'Filth and Foreign Material Inspection by Magnification': 'foreign_matter',
    'foreign_material_visual_inspection': 'foreign_matter',
    'FOREIGN MATERIALS': 'foreign_matter',
    'Foreign Materials': 'foreign_matter',
    'metal': 'heavy_metals',
    'MET': 'heavy_metals',
    'heavy_metals': 'heavy_metals',
    'Heavy Metals': 'heavy_metals',
    'Metals Analysis by ICPMS': 'heavy_metals',
    'Heavy Metals Screen': 'heavy_metals',
    'HEAVY METALS': 'heavy_metals',
    'microbial': 'microbes',
    'microbes': 'microbes',
    'microbiological': 'microbes',
    'microbiological_contaminants': 'microbes',
    'Microbiology': 'microbes',
    'microbiology': 'microbes',
    'Compliance Microbial': 'microbes',
    'Microbial Qpcr': 'microbes',
    'Microbial Result': 'microbes',
    'Microbial Screen': 'microbes',
    'MICROBIAL IMPURITIES': 'microbes',
    'Microbials by PCR': 'microbials',
    'moisture': 'moisture',
    'Moisture': 'moisture',
    'MOISTURE': 'moisture',
    'moisture_content': 'moisture',
    'Moisture Content': 'moisture',
    'Moisture by Moisture Balance': 'moisture',
    'Percent Moisture (%)': 'moisture',
    'moisture_analysis': 'moisture',
    'mycotoxins': 'mycotoxins',
    'MYCO': 'mycotoxins',
    'mycotoxin': 'mycotoxins',
    'Mycotoxin': 'mycotoxins',
    'Mycotoxins': 'mycotoxins',
    'Mycotoxins by LCMSMS': 'mycotoxins',
    'Mycotoxin Screen': 'mycotoxins',
    'MYCOTOXIN': 'mycotoxins',
    'PEST': 'pesticides',
    'pesticides': 'pesticides',
    'pesticide': 'pesticides',
    'Category 1 Pesticide': 'pesticides',
    'Category 2 Pesticide': 'pesticides',
    'Chemical Residue': 'pesticides',
    'Chemical Residue GC': 'pesticides',
    'Chemical Residue Gc': 'pesticides',
    'Pesticide Analysis by GCMS/LCMS': 'pesticides',
    'Pesticide Screen': 'pesticides',
    'Pesticide Screen Result - Category 1': 'pesticides',
    'Pesticide Screen Result - Category 2': 'pesticides',
    'PESTICIDES': 'pesticides',
    'solvent': 'residual_solvents',
    'RST': 'residual_solvents',
    'residual_solvents': 'residual_solvents',
    'residual_solvent': 'residual_solvents',
    'Residual Solvent Screen - Category 1': 'residual_solvents',
    'Residual Solvent Screen - Category 2': 'residual_solvents',
    'RESIDUAL SOLVENTS': 'residual_solvents',
    'terpenoids': 'terpenes',
    'TERP': 'terpenes',
    'terpenes': 'terpenes',
    'terpene': 'terpenes',
    'Terpenoid': 'terpenes',
    'terpenoid': 'terpenes',
    'Terpene': 'terpenes',
    'Terpene Analysis by GCMS': 'terpenes',
    'Terpene Test Result': 'terpenes',
    'TERPENES': 'terpenes',
    'WA': 'water_activity',
    'water_activity': 'water_activity',
    'Water Activity': 'water_activity',
    'Water Activity by Aqua Lab': 'water_activity',
    'Water Activity (Aw)': 'water_activity',
    'WATER ACTIVITY': 'water_activity',
    'terpenoid_testing': 'terpenes',
    # Necessary?
    'mycotoxins_status': 'mycotoxins',
    'pesticides_status': 'pesticides',
    'cannabinoids_status': 'cannabinoids',
    'heavy_metals_status': 'heavy_metals',
    'water_activity_status': 'water_activity',
    'Cannabinoid Profile': 'cannabinoids',
    'Microbiological Screen': 'microbes',
    'Mycotoxin Screen': 'mycotoxins',
}

# A map of encountered analytes to their standardized analyte,
# excluding analytes that can be standardized with `snake_case` alone.
# Optional: Create `analyte_details.json`.
ANALYTES = {
    'B1': 'aflatoxin_b1',
    'Aflatoxin B1': 'aflatoxin_b1',
    'B2': 'aflatoxin_b2',
    'Aflatoxin B2': 'aflatoxin_b2',
    'G1': 'aflatoxin_g1',
    'Aflatoxin G1': 'aflatoxin_g1',
    'G2': 'aflatoxin_g2',
    'Aflatoxin G2': 'aflatoxin_g2',
    'a_bisabolol': 'alpha_bisabolol',
    'a_cedrene': 'alpha_cedrene',
    'Ocimene isomer I': 'alpha_ocimene',
    'a_phellandrene': 'alpha_phellandrene',
    'a_pinene': 'alpha_pinene',
    'a_terpinene': 'alpha_terpinene',
    'alphaterpinene': 'alpha_terpinene',
    'A. flavus': 'aspergillus_flavus',
    'A.fumigatus': 'aspergillus_fumigatus',
    'a_fumigatus': 'aspergillus_fumigatus',
    'A. niger': 'aspergillus_niger',
    'a_niger': 'aspergillus_niger',
    'Aspergillus terreus': 'aspergillus_terreus',
    'A. terreus': 'aspergillus_terreus',
    'avermectin_b_1_a_abamectin': 'avermectin_b1b',
    'b_caryophyllene': 'beta_caryophyllene',
    'betacaryophyllene': 'beta_caryophyllene',
    'caryophyliene_oxide': 'caryophyllene_oxide',
    'Ocimene isomer II': 'beta_ocimene',
    'b_pinene': 'beta_pinene',
    'n_butane': 'butane',
    'Δ3-Carene': 'delta_3_carene',
    'Chlorantranil-iprole': 'chlorantraniliprole',
    'cis-Nerolidol': 'cis_nerolidol',
    'cy_uthrin': 'cyfluthrin',
    'delta_limonene': 'd_limonene',
    'limonene': 'd_limonene',
    'r_limonene': 'd_limonene',
    '(R)-(+)-Limonene': 'd_limonene',
    '3-Carene': 'delta_3_carene',
    '3_carene': 'delta_3_carene',
    '4_carene': 'delta_3_carene',
    'Δ8': 'delta_8_thc',
    'D8THC': 'delta_8_thc',
    'd_8_thc': 'delta_8_thc',
    'delta_8_thc_d_8_thc': 'delta_8_thc',
    'Δ9': 'delta_9_thc',
    'D9THC': 'delta_9_thc',
    'd_9_othc': 'delta_9_thc',
    'delta_9_thc_d_9_thc': 'delta_9_thc',
    'd_9_thca': 'thca',
    'd_9_thcva': 'thcva',
    'd_9_thcv': 'thcv',
    'cbca_3': 'cbca',
    'cbna_3': 'cbna',
    'cbl_3': 'cbl',
    'THCA-A': 'thca',
    'thca_a': 'thca',
    'ddvp_dichlorvous': 'dichlorvos',
    'Dichlorvos (DDVP)': 'dichlorvos',
    'dichlorvos_ddvp': 'dichlorvos',
    'DDVP (Dichlorvous)': 'dichlorvos',
    'ddvp_dichlorvos': 'dichlorvos',
    'SSCD': 'soil',
    'Escherichia Coli': 'e_coli',
    'escherichia_coli': 'e_coli',
    'Shiga toxin-producing E. coli': 'e_coli',
    'escherichia_coli_stec': 'e_coli',
    'escherichia_coli_shigella': 'e_coli',
    'STEC': 'e_coli',
    'IFM': 'foreign_matter',
    'Imbedded Foreign Material': 'foreign_matter',
    'IF RH ME': 'foreign_matter',
    'Insect Fragments, Hair, Mammal Excrement': 'foreign_matter',
    'g_terpinene': 'gamma_terpinene',
    'gama_terpinene': 'gamma_terpinene',
    'gammaterpinene': 'gamma_terpinene',
    'α-Humulene': 'alpha_humulene',
    'a_humulene': 'alpha_humulene',
    'alphahumulene': 'alpha_humulene',
    'humulene': 'alpha_humulene',
    'Moisture': 'moisture_content',
    'aflatoxin_b_1_b_2_g_1_g_2_and_ochratoxin_a': 'mycotoxins',
    'Naled': 'naled',
    'Ochratoxin A': 'ochratoxin_a',
    'β-Ocimene': 'beta_ocimene',
    'Oxamyl': 'oxamyl',
    'p-Cymene': 'p_cymene',
    'para_cymene': 'p_cymene',
    'p-Mentha-1,5-diene': 'p_mentha_1_5_diene',
    'p_mentha_15_diene': 'p_mentha_1_5_diene',
    'Piperonylbuto-xide': 'piperonyl_butoxide',
    'r_pulegone': 'pulegone',
    'Salmonella spp.': 'salmonella',
    'Salmonella SPP': 'salmonella',
    'Salmonella spp': 'salmonella',
    'salmonella_specific_gene': 'salmonella',
    'Sand, Soil, Cinders, Dirt': 'soil',
    'Sand': 'soil',
    'SscD': 'soil',
    'spinetoram_j': 'spinetoram',
    'spinetoram_l': 'spinetoram',
    'a_terpineol': 'alpha_terpineol',
    'Terpinen-4-ol': 'terpineol',
    'gammaterpineol': 'gamma_terpineol',
    'Aflatoxins': 'total_aflatoxins',
    'Total CBD(Total CBD = (CBDA x 0.877) + CBD)': 'total_cbd',
    'Total THC(Total THC = (THCA x 0.877) + THC)': 'total_thc',
    'trans_b_farnesene': 'trans_beta_farnesene',
    'trans_beta_famesene': 'trans_beta_farnesene',
    'Trifloxystrob-in': 'trifloxystrobin',
    'AW': 'water_activity',
    '* Beyond scope of accreditation': 'wildcard',
    'afalatoxin_b_2': 'aflatoxin_b2',
    'aflatoxin_b_1': 'aflatoxin_b1',
    'aflatoxin_b_2': 'aflatoxin_b2',
    'aflatoxin_g_1': 'aflatoxin_g1',
    'aflatoxin_g_2': 'aflatoxin_g2',
    'aspergillus_siger': 'aspergillus_niger',
    'avermectin_b_1_b_abamectin': 'avermectin_b1b',
    'b_ocimene': 'beta_ocimene',
    'myrcene': 'beta_myrcene',
    '0_1_guaiol': 'guaiol',
    '0_3_guaiol': 'guaiol',
    '1_2_dichloro_ethane': '1_2_dichloroethane',
    '8_thc': 'delta_8_thc',
    '9_thc': 'delta_9_thc',
    '9_thca': 'thca',
    'a_atoxin_b_1': 'aflatoxin_b1',
    'a_atoxin_g_1': 'aflatoxin_g1',
    'a_atoxin_g_2': 'aflatoxin_g2',
    'carene': 'delta_3_carene',
    'd_3_carene': 'delta_3_carene',
    'd_9_thc': 'delta_9_thc',
    'caryophyllene': 'beta_caryophyllene',
    'chlorantranilip_role': 'chlorantraniliprole',
    'trichloroethene': 'trichloroethylene',
    'xylenes': 'total_xylenes',
    'total_total_xylenes': 'total_xylenes',
    'total_sample_area_covered_by_mold': 'mold',
    'yeast_and_mold': 'mold',
    'total_sample_area_covered_by_sand_soil_cinders_or_dirt': 'soil',
    'total_sample_area_covered_by_an_imbedded_foreign_material': 'foreign_matter',
    'tri_oxystrobin': 'trifloxystrobin',
    'total_aflatoxin': 'total_aflatoxins',
    'spinosad': 'spinosad_a',
    'spinosyn_a_spinosad': 'spinosad_a',
    'spinosyn_d_spinosad': 'spinosad_d',
    'shiga_toxin_producing_escherichia_coli': 'e_coli',
    'sand_soils_cinders_and_dirt': 'soil',
    'salmonella_aoac': 'salmonella',
    'salmonella_spp': 'salmonella',
    'pyrethrins_pyrethrin_i': 'pyrethrins',
    'pyrethrins_pyrethrin_ii': 'pyrethrins',
    'pentachloroni_trobenzene': 'pentachloronitrobenzene',
    'pentachloronitro_benzene': 'pentachloronitrobenzene',
    'mammalian_excreta': 'foreign_matter',
    'mammalian_excreta_count': 'foreign_matter',
    'insect_fragment': 'foreign_matter',
    'insect_fragment_count': 'foreign_matter',
    'imbedded_foreign_material': 'foreign_matter',
    'hair': 'foreign_matter',
    'hair_count': 'foreign_matter',
    'loq_isoborneol': 'isoborneol',
    'loq_terpineol': 'terpineol',
    'l_fenchone': 'fenchone',
    'moisture': 'moisture_content',
    'piperonylbu_toxide': 'piperonyl_butoxide',
    'isopropanol': 'isopropyl_alcohol',
    '2_propanol': 'isopropyl_alcohol',
    'propan_2_ol': 'isopropyl_alcohol',
    'IPA': 'isopropyl_alcohol',
    'lsopropyl_alcohol': 'isopropyl_alcohol',
    'isopropylalcohol': 'isopropyl_alcohol',
    'methyl_parathion': 'parathion_methyl',
    'delta_9_thc_per_packagepackage_3_5_grams': 'total_thc_per_package',
    'n_hexane': 'hexane',
    'mevinphos_i': 'mevinphos',
    'mevinphos_ii': 'mevinphos',
    'permethrin_cis': 'permethrin',
    'permethrin_trans': 'permethrin',
    'permethrins': 'permethrin',
    'chlordane_cis': 'chlordane',
    'chlordane_trans': 'chlordane',
    'trans_geraniol': 'geraniol',
    'cis_geraniol': 'geraniol',
    'geranyl': 'geranyl_acetate',
    'geranylacetate': 'geranyl_acetate',
    'trans_ocimene': 'beta_ocimene',
    'cis_ocimene': 'alpha_ocimene',
    'farnesene': 'trans_beta_farnesene',
    'ethoprop_hos': 'ethoprophos',
    'fenpryoximate': 'fenpyroximate',
    'fludioxanil': 'fludioxonil',
    'thiaclorprid': 'thiacloprid',
    '9_thc_per_serving': 'delta_9_thc_per_serving',
    'd_9_thc_per_serving': 'delta_9_thc_per_serving',
    'n_pentane': 'pentane',
    'n_heptane': 'heptane',
    'carbotfuran': 'carbofuran',
    'lsobomneol': 'isoborneol',
    'lsoborneol': 'isoborneol',
    'lsopulegol': 'isopulegol',
    'methomy': 'methomyl',
    # FIXME: Standardize `fm`, `sscd`, `cep` without messing everything up.
    # 'ocimene': 'alpha_ocimene',
    'Tetrahydrocannabinolic acid': 'thca',
    'D9-Tetrahydrocannabinol': 'delta_9_thc',
    'Cannabidiolic acid': 'cbda',
    'Cannabidiol': 'cbd',
    'Cannabinol': 'cbn',
    'Cannabichromene': 'cbc',
    'Cannabigerolic acid': 'cbga',
    'Cannabigerol': 'cbg',
    'Cannabidivarin': 'cbdv',
    'Tetrahydrocannabivarin': 'thcv',
    'D8-Tetrahydrocannabinol': 'delta_8_thc',
    'Total Yeast and Mold': 'mold',
    'total_yeasttomold': 'mold',
    'Total Yeast/Mold': 'mold',
    'Bile-Tolerant Gram-Neg. Bacteria': 'enterobacteriaceae',
    'btgn': 'enterobacteriaceae',
    'Total Arsenic': 'arsenic',
    'Total Mercury': 'mercury',
    'arsenic_as': 'arsenic',
    'lead_pb': 'lead',
    'cadmium_cd': 'cadmium',
    'mercury_hg': 'mercury',
    'Arsenic (As)': 'arsenic',
    'Lead (Pb)': 'lead',
    'Cadmium (Cd)': 'cadmium',
    'Mercury (Hg)': 'mercury',
    'stec_e_coli': 'e_coli',
    'stec': 'e_coli',
    'escherichia_coli_specific': 'e_coli',
    'iso_butane': 'isobutane',
    'YM': 'mold',
    'CC': 'total_coliforms',
    'TAC': 'total_viable_aerobic_bacteria',
    'total_yeast_and_mold': 'mold',
    'total_bile_tolerant_gram_negative_bacteria': 'enterobacteriaceae',
    'mycotoxin': 'total_aflatoxins',
    'a_bisabolene': 'alpha_bisabolene',
    'a_bulnesene': 'alpha_bulnesene',
    'a_farnesene': 'alpha_farnesene',
    'a_maaliene': 'alpha_maaliene',
    'a_ocimene': 'alpha_ocimene',
    'a_thujone': 'alpha_thujone',
    'b_farnesene': 'trans_beta_farnesene',
    'b_maaliene': 'beta_maaliene',
    'bisabolol': 'alpha_bisabolol',
    'foreign_matter_i_h_e': 'foreign_matter',
}
# TODO: Find out how to map:
# cis_nerolidol vs. nerolidol vs. nerolidol_1 vs. nerolidol_2 vs. trans_nerolidol?

# A map of encountered fields to their standardized field.
# FIXME: Lab's sometimes use sample_id as lab_id.
# 'sample_id': 'lab_id',
STANDARD_FIELDS = {
    'Analyte': 'name',
    'Labeled Amount': 'sample_weight',
    'Limit': 'limit',
    'Detected': 'value',
    'LOD': 'lod',
    'LOQ': 'loq',
    'Pass/Fail': 'status',
    'metrc_src_uid': 'metrc_source_id',
    'matrix': 'product_type',
    'collected_on': 'date_collected',
    'received_on': 'date_received',
    'moisture': 'moisture_content',
    'terpenoids': 'terpenes',
    'foreign_materials': 'foreign_matter',
    'filth_and_foreign_matter': 'foreign_matter',
    'total_terpenoids_mgtog': 'total_terpenes_mg_g',
    'Source Batch ID': 'batch_id',
    'batch_number': 'batch_number',
    'batch_size': 'batch_size',
    '9_thc_per_unit': 'cannabinoids_status',
    'collected': 'date_collected',
    'Collected': 'date_collected',
    'harvesttoprocessing_date': 'date_harvested',
    'Harvest/Processing Date': 'date_harvested',
    'manufacture_date': 'date_produced',
    'received': 'date_received',
    'date_received': 'date_received',
    'date_sampled': 'date_sampled',
    'Date': 'date_tested',
    'date_reported': 'date_tested',
    'dates_of_analysis': 'date_tested',
    'business_name': 'distributor',
    'license_number': 'distributor_license_number',
    'foreign_material_method': 'foreign_matter_method',
    'foreign_material': 'foreign_matter_status',
    'heavy_metals': 'heavy_metals_status',
    'Instrument': 'instrument',
    'coa_id': 'lab_id',
    'sample_no': 'lab_id',
    'lab_sample_id': 'lab_id',
    'Lab Sample ID': 'lab_id',
    'sample_code': 'lab_id',
    'Lab ID': 'lab_id',
    'licensenumber': 'lab_license_number',
    'action-limit': 'limit',
    'action': 'limit',
    'limit': 'limit',
    'LOD mg/g': 'lod',
    'LOD (mg/g)': 'lod',
    'LOD (ug/g)': 'lod',
    'LOD (ug/kg)': 'lod',
    'LOD/LOQ': 'lod',
    'lod': 'lod',
    'LOQ mg/g': 'loq',
    'LOQ (mg/g)': 'loq',
    'LOQ (ug/g)': 'loq',
    'LOQ (ug/kg)': 'loq',
    'loq': 'loq',
    'mu': 'margin_of_error',
    'sample_matrix': 'matrix',
    'Method': 'method',
    'track_and_trace_test_package': 'metrc_lab_id',
    'Test RFID': 'metrc_lab_id',
    'Lab Metrc Id': 'metrc_lab_id',
    'source_metrc_uid': 'metrc_source_id',
    'metrc_uid': 'metrc_source_id',
    'batch_id': 'metrc_source_id',
    'track_and_trace_source_package_s': 'metrc_source_id',
    'Source RFID': 'metrc_source_id',
    'metrc_tag': 'metrc_source_id',
    'Client Metrc Id': 'metrc_source_id',
    'mg/g': 'mg_g',
    'result-mass': 'mg_g',
    'microbiology_status': 'microbes_status',
    'microbial_status': 'microbes_status',
    'microbials_status': 'microbes_status',
    'microbiology': 'microbes_status',
    'mycotoxins': 'mycotoxins_status',
    'Target Analyte': 'name',
    'Microbiological Assay': 'name',
    'compound': 'name',
    'Cannabinoid': 'name',
    'analyte': 'name',
    'name': 'name',
    'Order Number': 'order_number',
    'pesticides': 'pesticides_status',
    'Company': 'producer',
    'client': 'producer',
    'client_address': 'producer_address',
    'client_license_number': 'producer_license_number',
    'sample_name': 'product_name',
    'Sample Name': 'product_name',
    'unit_mass': 'product_size',
    'Type': 'product_type',
    'Matrix': 'product_type',
    'sample_type': 'product_type',
    'Matrix Type': 'product_type',
    'Order ID': 'project_id',
    'residual_solvents': 'residual_solvents_status',
    'sample_size': 'sample_size',
    'Sample Wt': 'sample_weight',
    'total_sample_weight_g': 'sample_weight',
    'product_density': 'sample_weight',
    'Product Density': 'sample_weight',
    'primary_sample': 'sample_weight',
    'Density': 'sample_weight',
    'increment_g': 'sample_weight_used',
    'sampling_method': 'sampling_method',
    'Serving Size': 'serving_size',
    'Servings': 'servings_per_package',
    'overall': 'status',
    'overall_batch': 'status',
    'result-pf': 'status',
    'Status': 'status',
    'level': 'status',
    'status': 'status',
    'sum_of_cannabinoids': 'sum_of_cannabinoids',
    'total_cbc': 'total_cbc',
    'total_cbd': 'total_cbd',
    'total_cbdv': 'total_cbdv',
    'total_cbg': 'total_cbg',
    'total_terpenoids': 'total_terpenes',
    'total_terpenoids_percent': 'total_terpenes',
    'Total Terpenes (%)': 'total_terpenes',
    'total_thc': 'total_thc',
    'total_thcv': 'total_thcv',
    '% Test': 'value',
    'PPM': 'value',
    'PPM (ug/g)': 'value',
    'PPM (ug/kg)': 'value',
    'PPB (ug/g)': 'value',
    'PPB (ug/kg)': 'value',
    'Threshold': 'value',
    'result-percent': 'value',
    'Findings': 'value',
    '%': 'value',
    'result': 'value',
    'value': 'value',
    'cannabinoids': 'cannabinoids_status',
    'water_activity': 'water_activity_status',
    'terpene_analysis_add_on': 'terpenes_status',
    'microbials': 'microbials_status',
    'metals': 'heavy_metals_status',
    'Sampling Method/SOP': 'sampling_method',
    'Batch Size': 'batch_size',
    'Sample Size': 'sample_size',
    'Date Sampled': 'date_sampled',
    'Date Received': 'date_received',
    'total_batch': 'batch_size',
    'total_cannabinoids': 'total_cannabinoids',
    'Total Cannabinoids (%)': 'total_cannabinoids',
    'Total Calculated d9-THC (%)': 'total_thc',
    'Total Calculated CBD (%)': 'total_cbd',
    'batch_size_sample_size': 'batch_size',
    'collected_received': 'date_collected',
    'lloq': 'loq',
    'Anresco ID': 'lab_id',
    'Lot/Batch Number': 'batch_number',
    'cannabinoid_pro_le': 'cannabinoids_status',
    'pesticide_residue_screen': 'pesticides_status',
    'microbiological_screen': 'microbes_status',
    'heavy_metal_screen': 'heavy_metals_status',
    'mycotoxin_screen': 'mycotoxins_status',
    'other_analyses': 'moisture_status',
    'total_active_cannabinoids': 'total_cannabinoids',
    'batch_no': 'batch_number',
    'batch_size_lbs': 'batch_size',
    'cannabinoid_method': 'cannabinoids_method',
    'foreign_matter_status_method': 'foreign_matter_method',
    'heavy_metal_method': 'heavy_metals_method',
    'heavy_metals_status_status': 'heavy_metals_status',
    'moisture_content_method': 'moisture_method',
    'mycotoxin_method': 'mycotoxins_method',
    'mycotoxins_status_status': 'mycotoxins_status',
    'pesticide_method': 'pesticides_method',
    'pesticides_status_status': 'pesticides_status',
    'terpenoid_method': 'terpenes_method',
    'total_a_atoxins': 'total_aflatoxins',
    'water_activity_status_method': 'water_activity_method',
    'water_activity_status_status': 'water_activity_status',
    'moisture_content_status': 'moisture_status',
    'percent_moisture_status': 'moisture_status',
    'total_d_9_thc_mg_per_package': 'total_thc_mg_per_package'
}

# A map of standard units by analysis to use when no units are obtainable.
STANDARD_UNITS = {
    'cannabinoids': 'percent',
    'foreign_matter': 'percent',
    'heavy_metals': 'μg/g',
    'microbes': 'CFU/g',
    'moisture': 'percent',
    'mycotoxins': 'μg/g',
    'pesticides': 'μg/g',
    'terpenes': 'percent',
    'water_activity': 'aW',
}

# A map of encountered product types to their standardized product type.
PRODUCT_TYPES = {
    'Flower & Buds': 'flower',
    'Immature Plants': 'immature_plant',
    'Concentrate (Non-Solvent Based) (Count-Volume)': 'non_solvent_concentrate',
    'Concentrate (Non-Solvent Based) (Count-Weight)': 'non_solvent_concentrate',
    'Concentrate (Weight Based)': 'concentrate',
    'Edibles (Count-Volume)': 'solid_edible',
    'Edibles (Count-Weight)': 'solid_edible',
    'Extracts (Solvent Based) (Count-Volume)': 'concentrate',
    'Extracts (Solvent Based) (Count-Weight)': 'concentrate',
    'Kief': 'kief',
    'Mature Plants': 'mature_plant',
    'Metered Dose Nasal Spray Products': 'nasal_spray',
    'MMJ Waste': 'waste',
    'Pre-Roll (Flower Only)': 'pre_roll',
    'Pre-Roll (Infused)': 'infused_pre_roll',
    'Pressurized Metered Dose Inhaler Products': 'nasal_spray',
    'Rectal/Vaginal Administration Products (Count-Volume)': 'suppository',
    'Rectal/Vaginal Administration Products (Count-Weight)': 'suppository',
    'Seeds': 'seeds',
    'Shake/Trim': 'shake',
    'Shake/Trim (by Strain)': 'shake',
    'Tinctures (Count-Volume)': 'tincture',
    'Tinctures (Count-Weight)': 'tincture',
    'Topicals (Count-Volume)': 'topical',
    'Topicals (Count-Weight)': 'topical',
    'Transdermal Patches': 'transdermal',
    'Vape Cartridges': 'vape_cartridge',
    'Whole Wet Plant': 'plant',
    'Buds': 'flower',
    'Infused': 'solid_edible',
    'InfusedEdible': 'solid_edible',
    'Infused Liquid': 'liquid_edible',
}

# A map of encountered strains to their standardized strain name.
STRAINS = {
    'Cannatonic #4': 'Cannatonic',
    'Cannatonic #4 - RP': 'Cannatonic',
    'GSC': 'Girl Scout Cookies',
    'Gorilla Glue #4': 'Gorilla Glue',
    'GG': 'Gorilla Glue',
    'GG4': 'Gorilla Glue',
    'GG #4': 'Gorilla Glue',
    'GG#4': 'Gorilla Glue',
    'GG 4': 'Gorilla Glue',
    'Gorilla Glue 4': 'Gorilla Glue',
    'Gorilla Glue #4 - 1': 'Gorilla Glue',
    'GDP': 'Grand Daddy Purple',
    'Granddaddy Purple': 'Grand Daddy Purple',
    'SFV': 'SFV OG',
}

# Standard value codings.
CODINGS = {
    'ND': 0.000000001,
    'No detection in 1 gram': 0.000000001,
    'Negative/1g': 0.000000001,
    'PASS': 0.000000001,
    'LOD': 0.00000001,
    '<LOD': 0.00000001,
    '< LOD': 0.00000001,
    '<LOQ': 0.0000001,
    '< LOQ': 0.0000001,
    '<LLoQ': 0.0000001,
    '<LLOQ': 0.0000001,
    'BLQ': 0.0000001,
    '<LLoa': 0.0000001,
    '≥ LOD': 10_001,
    'NR': None,
    'N/A': None,
    'na': None,
    'NT': None,
}

# Cannabinoid decarboxylation rate.
DECARB = 0.877 # Source: <https://www.conflabs.com/why-0-877/>

# Default headers to use for HTTP requests, because we are AI and should not be treated as a bot.
DEFAULT_HEADERS = {'User-Agent': 'Mozilla/5.0 (Windows NT 6.2; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/32.0.1667.0 Safari/537.36'}

# Random characters to use in password generation.
RANDOM_STRING_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

# A map of state abbreviations to state names.
states = {
    'AL': 'Alabama',
    'AK': 'Alaska',
    'AZ': 'Arizona',
    'AR': 'Arkansas',
    'CA': 'California',
    'CO': 'Colorado',
    'CT': 'Connecticut',
    'DE': 'Delaware',
    'FL': 'Florida',
    'GA': 'Georgia',
    'HI': 'Hawaii',
    'ID': 'Idaho',
    'IL': 'Illinois',
    'IN': 'Indiana',
    'IA': 'Iowa',
    'KS': 'Kansas',
    'KY': 'Kentucky',
    'LA': 'Louisiana',
    'ME': 'Maine',
    'MD': 'Maryland',
    'MA': 'Massachusetts',
    'MI': 'Michigan',
    'MN': 'Minnesota',
    'MS': 'Mississippi',
    'MO': 'Missouri',
    'MT': 'Montana',
    'NE': 'Nebraska',
    'NV': 'Nevada',
    'NH': 'New Hampshire',
    'NJ': 'New Jersey',
    'NM': 'New Mexico',
    'NY': 'New York',
    'NC': 'North Carolina',
    'ND': 'North Dakota',
    'OH': 'Ohio',
    'OK': 'Oklahoma',
    'OR': 'Oregon',
    'PA': 'Pennsylvania',
    'RI': 'Rhode Island',
    'SC': 'South Carolina',
    'SD': 'South Dakota',
    'TN': 'Tennessee',
    'TX': 'Texas',
    'UT': 'Utah',
    'VT': 'Vermont',
    'VA': 'Virginia',
    'WA': 'Washington',
    'WV': 'West Virginia',
    'WI': 'Wisconsin',
    'WY': 'Wyoming',
    'DC': 'District of Columbia',
}

# A map of state names to state abbreviations.
state_names = {
    'Alabama': 'AL',
    'Alaska': 'AK',
    'Arizona': 'AZ',
    'Arkansas': 'AR',
    'California': 'CA',
    'Colorado': 'CO',
    'Connecticut': 'CT',
    'Delaware': 'DE',
    'Florida': 'FL',
    'Georgia': 'GA',
    'Hawaii': 'HI',
    'Idaho': 'ID',
    'Illinois': 'IL',
    'Indiana': 'IN',
    'Iowa': 'IA',
    'Kansas': 'KS',
    'Kentucky': 'KY',
    'Louisiana': 'LA',
    'Maine': 'ME',
    'Maryland': 'MD',
    'Massachusetts': 'MA',
    'Michigan': 'MI',
    'Minnesota': 'MN',
    'Mississippi': 'MS',
    'Missouri': 'MO',
    'Montana': 'MT',
    'Nebraska': 'NE',
    'Nevada': 'NV',
    'New Hampshire': 'NH',
    'New Jersey': 'NJ',
    'New Mexico': 'NM',
    'New York': 'NY',
    'North Carolina': 'NC',
    'North Dakota': 'ND',
    'Ohio': 'OH',
    'Oklahoma': 'OK',
    'Oregon': 'OR',
    'Pennsylvania': 'PA',
    'Rhode Island': 'RI',
    'South Carolina': 'SC',
    'South Dakota': 'SD',
    'Tennessee': 'TN',
    'Texas': 'TX',
    'Utah': 'UT',
    'Vermont': 'VT',
    'Virginia': 'VA',
    'Washington': 'WA',
    'West Virginia': 'WV',
    'Wisconsin': 'WI',
    'Wyoming': 'WY',
    'District of Columbia': 'DC',
    'American Samoa': 'AS',
    'Guam': 'GU',
    'Northern Mariana Islands': 'MP',
    'Puerto Rico': 'PR',
    'United States Minor Outlying Islands': 'UM',
    'U.S. Virgin Islands': 'VI'
}

# A map of state abbreviations to timezone.
state_time_zones = {
    'AL': 'America/Chicago',
    'AK': 'America/Anchorage',
    'AZ': 'America/Phoenix',
    'AR': 'America/Chicago',
    'CA': 'America/Los_Angeles',
    'CO': 'America/Denver',
    'CT': 'America/New_York',
    'DC': 'America/New_York',
    'DE': 'America/New_York',
    'FL': 'America/New_York',
    'GA': 'America/New_York',
    'HI': 'Pacific/Honolulu',
    'ID': 'America/Denver',
    'IL': 'America/Chicago',
    'IN': 'America/Indiana/Indianapolis',
    'IA': 'America/Chicago',
    'KS': 'America/Chicago',
    'KY': 'America/New_York',
    'LA': 'America/Chicago',
    'ME': 'America/New_York',
    'MD': 'America/New_York',
    'MA': 'America/New_York',
    'MI': 'America/New_York',
    'MN': 'America/Chicago',
    'MS': 'America/Chicago',
    'MO': 'America/Chicago',
    'MT': 'America/Denver',
    'NE': 'America/Chicago',
    'NV': 'America/Los_Angeles',
    'NH': 'America/New_York',
    'NJ': 'America/New_York',
    'NM': 'America/Denver',
    'NY': 'America/New_York',
    'NC': 'America/New_York',
    'ND': 'America/North_Dakota/Center',
    'OH': 'America/New_York',
    'OK': 'America/Chicago',
    'OR': 'America/Los_Angeles',
    'PA': 'America/New_York',
    'RI': 'America/New_York',
    'SC': 'America/New_York',
    'SD': 'America/Chicago',
    'TN': 'America/Chicago',
    'TX': 'America/Chicago',
    'UT': 'America/Denver',
    'VT': 'America/New_York',
    'VA': 'America/New_York',
    'WA': 'America/Los_Angeles',
    'WV': 'America/New_York',
    'WI': 'America/Chicago',
    'WY': 'America/Denver',
}


if __name__ == '__main__':

    from cannlytics.utils.utils import snake_case

    # Sort a given dictionary and remove duplicate fields.
    # Also Add snake_case key if not already present.
    # This is useful when adding new fields to a constant in development.
    x = {}
    for k, v in STANDARD_FIELDS.items():
        if k not in x.keys():
            x[k] = v
            x[snake_case(k)] = v
    x = {k: v for k, v in sorted(x.items(), key=lambda item: item[1])}
    print(x)



"""
Project Management | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 11/2/2021  
Updated: 11/2/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Manage your laboratory projects.
"""

# TODO: Implement various project helper functions here.



"""
COA AI | CoADoc
Copyright (c) 2021-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 6/12/2023
Updated: 7/4/2023
License: MIT License <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse COAs with the aid of OpenAI's API.

"""
# Standard imports:
from datetime import datetime
import os
import json
import tempfile
from time import sleep
from typing import Any, Optional

# External imports:
from dotenv import dotenv_values
import openai
import pandas as pd
import pdfplumber
import requests

# Internal imports:
from cannlytics import __version__
from cannlytics.ai import (
    AI_WARNING,
    INSTRUCTIONAL_PROMPT,
    initialize_openai,
    get_messages_price,
    split_into_token_chunks,
)
from cannlytics.data import create_hash, create_sample_id
from cannlytics.utils import snake_case
from cannlytics.utils.constants import DEFAULT_HEADERS


#-----------------------------------------------------------------------
# First, try to get the metadata from the front page.
#-----------------------------------------------------------------------

# Prompt to parse metadata from the first page.
COA_PROMPT = """Given text, extract JSON, where:

| Field | Example | Description |
|-------|---------|-------------|
| `analyses` | ["cannabinoids"] | A list of analyses performed on a given sample. |
| `{analysis}_status` | "pass" | The pass, fail, or N/A status for pass / fail analyses.   |
| `methods` | [{"analysis: "cannabinoids", "method": "HPLC"}] | The methods used for each analysis. |
| `date_collected` | 2022-04-20T04:20 | An ISO-formatted time when the sample was collected. |
| `date_tested` | 2022-04-20T16:20 | An ISO-formatted time when the sample was tested. |
| `date_received` | 2022-04-20T12:20 | An ISO-formatted time when the sample was received. |
| `lab` | "MCR Labs" | The lab that tested the sample. |
| `lab_address` | "85 Speen St, Framingham, MA 01701" | The lab's address. |
| `lab_street` | "85 Speen St" | The lab's street. |
| `lab_city` | "Framingham" | The lab's city. |
| `lab_state` | "MA" | The lab's state. |
| `lab_zipcode` | "01701" | The lab's zipcode. |
| `distributor` | "Fred's Dispensary" | The name of the product distributor, if applicable. |
| `distributor_address` | "420 State Ave, Olympia, WA 98506" | The distributor address, if applicable. |
| `distributor_street` | "420 State Ave" | The distributor street, if applicable. |
| `distributor_city` | "Olympia" | The distributor city, if applicable. |
| `distributor_state` | "WA" | The distributor state, if applicable. |
| `distributor_zipcode` | "98506" | The distributor zip code, if applicable. |
| `distributor_license_number` | "L-123" | The distributor license number, if applicable. |
| `producer` | "Grow House" | The producer of the sampled product. |
| `producer_address` | "3rd & Army, San Francisco, CA 55555" | The producer's address. |
| `producer_street` | "3rd & Army" | The producer's street. |
| `producer_city` | "San Francisco" | The producer's city. |
| `producer_state` | "CA" | The producer's state. |
| `producer_zipcode` | "55555" | The producer's zipcode. |
| `producer_license_number` | "L2Calc" | The producer's license number. |
| `product_name` | "Blue Rhino Pre-Roll" | The name of the product. |
| `lab_id` | "Sample-0001" | A lab-specific ID for the sample. |
| `product_type` | "flower" | The type of product. |
| `batch_number` | "Order-0001" | A batch number for the sample or product. |
| `traceability_ids` | ["1A4060300002199000003445"] | A list of relevant traceability IDs. |
| `product_size` | 2000 | The size of the product in milligrams. |
| `serving_size` | 1000 | An estimated serving size in milligrams. |
| `servings_per_package` | 2 | The number of servings per package. |
| `sample_weight` | 1 | The weight of the product sample in grams. |
| `status` | "pass" | The overall pass / fail status for all contaminant screening analyses. |
| `total_cannabinoids` | 14.20 | The analytical total of all cannabinoids measured. |
| `total_thc` | 14.00 | The analytical total of THC and THCA. |
| `total_cbd` | 0.20 | The analytical total of CBD and CBDA. |
| `total_terpenes` | 0.42 | The sum of all terpenes measured. |
| `sample_id` | "{sha256-hash}" | A generated ID to uniquely identify the `producer`, `product_name`, and `date_tested`. |
| `strain_name` | "Blue Rhino" | A strain name, if specified. Otherwise, can be attempted to be parsed from the `product_name`. |
"""

# Prompt used to parse results from all pages.
RESULTS_PROMPT = """Given text, extract only results as a list of JSON objects:

{
    "results": [
        {
            "analysis": str,
            "key": str,
            "name": str,
            "value": float,
            "mg_g": float,
            "units": str,
            "limit": float,
            "lod": float,
            "loq": float,
            "status": str
        }
    ]

}

Where:

| Field | Example| Description |
|-------|--------|-------------|
| `analysis` | "pesticides" | The analysis used to obtain the result. |
| `key` | "pyrethrins" | A standardized key for the result analyte. |
| `name` | "Pyrethrins" | The lab's internal name for the result analyte |
| `value` | 0.42 | The value of the result. |
| `mg_g` | 0.00000042 | The value of the result in milligrams per gram. |
| `units` | "ug/g" | The units for the result `value`, `limit`, `lod`, and `loq`. |
| `limit` | 0.5 | A pass / fail threshold for contaminant screening analyses. |
| `lod` | 0.01 | The limit of detection for the result analyte. Values below the `lod` are typically reported as `ND`. |
| `loq` | 0.1 | The limit of quantification for the result analyte. Values above the `lod` but below the `loq` are typically reported as `<LOQ`. |
| `status` | "pass" | The pass / fail status for contaminant screening analyses. |
"""

def parse_coa_with_ai(
        parser: Any,
        doc: str,
        temp_path: Optional[str] = None,
        session: Optional[Any] = None,
        headers: Optional[dict] = DEFAULT_HEADERS,
        use_cached: Optional[bool] = False,
        openai_api_key: Optional[str] = None,
        model: Optional[str] = 'gpt-4',
        max_tokens: Optional[int] = 4_000,
        temperature: Optional[float] = 0.0,
        initial_cost: Optional[float] = 0.0,
        instructional_prompt: Optional[str] = None,
        results_prompt: Optional[str] = None,
        coa_prompt: Optional[str] = None,
        max_prompt_length: Optional[int] = 4_000,
        verbose: Optional[bool] = False,
        user: Optional[str] = None,
        retry_pause: Optional[float] = 3.33,
    ) -> dict:
    """Parse a COA with OpenAI's GPT model and return the data as JSON."""

    # === DEV ===
    # from cannlytics.data.coas.coas import CoADoc
    # parser = CoADoc()
    # model = 'gpt-4'
    # max_tokens = 4_000
    # temperature = 0.0
    # initial_cost = 0.0
    # use_cached = True
    # instructional_prompt = INSTRUCTIONAL_PROMPT
    # coa_prompt = COA_PROMPT
    # results_prompt = RESULTS_PROMPT
    # max_prompt_length = 1_000
    # verbose = True
    # user = 'cannlytics'
    # doc = '../../../tests/assets/coas/gtl/Pineapple-XX-5-13-2129146.pdf'

    # Initialize prompts.
    if instructional_prompt is None:
        instructional_prompt = INSTRUCTIONAL_PROMPT
    if coa_prompt is None:
        coa_prompt = COA_PROMPT
    if results_prompt is None:
        results_prompt = RESULTS_PROMPT

    # Parse an observation.
    obs = {}

    # Track costs and prompts.
    cost = initial_cost
    prompts = []

    # If the `doc` is a URL, then download the PDF to the `temp_path`.
    # Then use the path of the downloaded PDF as the doc.
    coa_url = None
    if isinstance(doc, str):
        if doc.startswith('https'):
            coa_url = doc
            if temp_path is None: temp_path = tempfile.gettempdir()
            if not os.path.exists(temp_path): os.makedirs(temp_path)
            try:
                filename = doc.split('/')[-1].split('?')[0] + '.pdf'
            except:
                filename = 'coa.pdf'
            coa_pdf = os.path.join(temp_path, filename)
            if session is not None:
                response = session.get(doc)
            else:
                response = requests.get(doc, headers=headers)
            with open(coa_pdf, 'wb') as pdf:
                pdf.write(response.content)
            report = pdfplumber.open(coa_pdf)
            obs['coa_pdf'] = filename
        else:
            report = pdfplumber.open(doc)
            obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]

    # Get the text of the PDF.
    report = pdfplumber.open(doc)
    front_page_text = report.pages[0].extract_text()
    all_text = '\n\n'.join([page.extract_text() for page in report.pages])

    # Record the COA URL.
    if coa_url is None:
        coa_url = parser.find_pdf_qr_code_url(report)
    if coa_url is not None:
        filename = coa_url.split('/')[-1].split('?')[0] + '.pdf'
        obs['coa_urls'] = json.dumps([{'url': coa_url, 'filename': filename}])
        obs['lab_results_url'] = coa_url
    else:
        obs['coa_urls'] = None
        obs['lab_results_url'] = None

    # TODO: Get images from the PDF.
    images = []

    # Close the report.
    report.close()

    # See if the PDF or URL has already been parsed.
    # by checking if the hash exists in the database.
    coa_hash = create_hash(all_text, private_key='')
    obs['coa_hash'] = coa_hash
    # if use_cached:
    #     try:
    #         database = initialize_firebase()
    #         results = get_collection(
    #             'public/data/lab_results',
    #             filters=[{'key': 'coa_hash', 'operation': '==', 'value': coa_hash}],
    #             limit=1,
    #             database=database,
    #         )
    #         if results:
    #             return results[0], [], 0
    #     except:
    #         pass

    # Format the prompt.
    metadata_prompt = 'Text: ' + front_page_text + '\n\nJSON:'
    try:
        messages = [
            {'role': 'system', 'content': coa_prompt},
            {'role': 'system', 'content': instructional_prompt},
            {'role': 'user', 'content': metadata_prompt},
        ]
        cost += get_messages_price(messages, model=model)
        if verbose:
            print('MESSAGES:', messages)
        try:
            initialize_openai(openai_api_key)
            metadata_response = openai.ChatCompletion.create(
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                user=user,
            )
        except:
            if retry_pause:
                sleep(retry_pause)
                initialize_openai(openai_api_key)
                metadata_response = openai.ChatCompletion.create(
                    model=model,
                    messages=messages,
                    max_tokens=max_tokens,
                    temperature=temperature,
                    user=user,
                )
        prompts.append({
            'messages': messages,
            'completion': json.dumps(metadata_response.to_dict()),
        })
        content = metadata_response['choices'][0]['message']['content']
        if verbose:
            print('CONTENT:', content)
    except:
        if verbose:
            print('Parse metadata OpenAI query failed.')
    
    # Get the structured the data.
    try:
        start_index = content.find('{')
        end_index = content.rfind('}') + 1
        metadata = json.loads(content[start_index:end_index])
        obs = {**obs, **metadata}
    except:
        if verbose:
            print('Metadata JSON parsing failed.')

    #-----------------------------------------------------------------------
    # Second, try to get results from each page.
    #-----------------------------------------------------------------------

    # Extract the results
    results = []

    # Split the long string into smaller strings.
    # Optional: Try to compress the data.
    # substrings = split_string(all_text, max_prompt_length - round(0.1 * max_prompt_length))
    substrings = split_into_token_chunks(all_text, max_prompt_length)

    # Format the message.
    # content = 'Text: ' + all_text + '\n\nList of results as JSON:'
    for substring in substrings:
        messages = [
            {'role': 'system', 'content': results_prompt},
            {'role': 'system', 'content': instructional_prompt},
            # {'role': 'system', 'content': content},
        ]
        content = 'Text: ' + substring + '\n\nList of results as JSON:'
        messages.append({'role': 'user', 'content': content})

        # TODO: Try compression.
        # json_data = json.dumps(obs)
        # compressed_data = zlib.compress(json_data.encode())

        try:
            if verbose:
                print('MESSAGES:', messages)
            cost += get_messages_price(messages, model=model)
            initialize_openai(openai_api_key)
            try:
                initialize_openai(openai_api_key)
                results_response = openai.ChatCompletion.create(
                    model=model,
                    messages=messages,
                    max_tokens=max_tokens,
                    temperature=temperature,
                    user=user,
                )
            except:
                if retry_pause:
                    sleep(retry_pause)
                    initialize_openai(openai_api_key)
                    results_response = openai.ChatCompletion.create(
                        model=model,
                        messages=messages,
                        max_tokens=max_tokens,
                        temperature=temperature,
                        user=user,
                    )
            prompts.append({
                'messages': messages,
                'completion': json.dumps(results_response.to_dict()),
            })
            for choice in results_response['choices']:
                content = choice['message']['content']
                if verbose:
                    print('CONTENT:', content)
                start_index = content.find('{')
                end_index = content.rfind('}') + 1
                try:
                    results.extend(json.loads(content[start_index:end_index])['results'])
                except:
                    if verbose:
                        print('Failed to extend results.')
        except:
            if verbose:
                print('Parse results OpenAI query failed.')

    #-----------------------------------------------------------------------
    # Finally, combine and standardize the data,
    # warning users that the data was generated by AI.
    # Optional: If parsing does not work well, break up parsing requests.
    #-----------------------------------------------------------------------

    # Standardize analyses and methods.
    # TODO: Standardize analyses names.
    try:
        obs['analyses'] = json.dumps(list(set(obs['analyses'])))
    except:
        pass
    try:
        obs['methods'] = json.dumps(list(set(obs['methods'])))
    except:
        pass

    # TODO: Lookup additional details.
    # - lab_phone
    # - lab_email
    # - lab_image_url
    # - lab_county
    # - lab_latitude
    # - lab_longitude

    # TODO: Standardize results.
    for i, result in enumerate(results):
        key = snake_case(result['name'])
        results[i]['key'] = parser.analytes.get(key, key)

    # Standardize dates.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Standardize the data.
    obs['coa_algorithm'] = 'coa_ai.py'
    obs['coa_algorithm_entry_point'] = 'parse_coa_with_ai'
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['images'] = json.dumps(images)
    obs['results'] = json.dumps(results)
    obs['results_hash'] = create_hash(obs['results'])
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs.get('product_name', 'Unknown'),
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    obs['sample_hash'] = create_hash(obs)
    obs['warning'] = AI_WARNING

    # Return the data.
    return obs, prompts, cost


# === Tests ===
# Performed 2023-06-12 by Keegan Skeate <admin@cannlytics.com>.
if __name__ == '__main__':

    # Initialize CoADoc
    from cannlytics.data.coas.coas import CoADoc

    # Initialize OpenAI.
    config = dotenv_values('../../../.env')
    openai_api_key = config['OPENAI_API_KEY']
    
    # [✓] TEST: Parse a COA with AI.
    doc = '../../../tests/assets/coas/gtl/Pineapple-XX-5-13-2129146.pdf'
    parser = CoADoc()
    data, prompts, cost = parse_coa_with_ai(
        parser,
        doc,
        openai_api_key=openai_api_key,
    )
    assert data is not None
    print(data)

    # Save the data.
    data['cost'] = cost
    outfile = '../../../tests/assets/coas/gtl/Pineapple-XX-5-13-2129146.json'
    with open(outfile, 'w') as f:
        json.dump(data, f, indent=4)



"""
Worksheets | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 7/18/2021
Updated: 1/10/2022
License: MIT License <https://opensource.org/licenses/MIT>
"""
# Standard packages
from ast import literal_eval
from typing import Any, List, Optional, Tuple, Union

# External packages.
from dotenv import dotenv_values
from pandas import DataFrame, to_datetime
import requests
try:
    import xlwings
    from xlwings.utils import rgb_to_int
except ImportError:
    # FIXME: Work on alternatives as xlwings doesn't work in App Engine.
    pass

# Internal packages.
from ..utils.utils import snake_case


def format_values(data: Union[list, dict], columns: List[str]) -> List[dict]:
    """Format response data by given columns. Handles a list or dictionary of data.
    Args:
        data (list,dict): A list of values or a dictionary of values.
        columns (list): A list of column names used to get values from the data.
    Returns:
        (list): Returns a list of items.
    """
    items = []
    try:
        for item in data:
            values = []
            for column in columns:
                values.append(item.get(column))
            items.append(values)
    except AttributeError:
        values = []
        for column in columns:
            values.append(data.get(column))
        items = [values]
    return items


def get_data_block(sheet: Any, coords: str, expand: Optional[str] = None) -> dict:
    """Get a data block.
    Args:
        sheet (Sheet): The worksheet containing the data block.
        coords (str): The inclusive coordinates of the data block.
        expand (str): Optionally expand the range of values.
    Returns
        (dict): A dictionary of the data in the data block.
    """
    data = {}
    values = sheet.range(coords).options(expand=expand).value
    for item in values:
        key = snake_case(item[0])
        value = item[1]
        data[key] = value
    return data


def get_data_model(worksheet: Any, config: dict, model_type: str, ids: List[str]) -> Any:
    """Get a specific data model from the Cannlytics API.
    Args:
        worksheet (Worksheet): An xlwings Excel worksheet instance.
        config (dict): A dictionary of user configuration.
        model_type (str): The data model name.
        ids (list): Specific IDs to retrieve from the API.
    Returns:
        (HTTPResponse): An HTTP response from the Cannlytics API.
    """
    base = config['api_url']
    org_id = worksheet.range(config['org_id_cell']).value
    env_variables = dotenv_values(config['env_path'])
    api_key = env_variables['CANNLYTICS_API_KEY']
    headers = {
        'Authorization': 'Bearer %s' % api_key,
        'Content-type': 'application/json',
    }
    if len(ids) == 1:
        url = f'{base}/{model_type}/{ids[0]}?organization_id={org_id}'
    else:
        url = f'{base}/{model_type}?organization_id={org_id}&items={str(ids)}'
    response = requests.get(url, headers=headers)
    return response


def increment_row(coords: str) -> str:
    """Increment a row given its starting coordinates.
    Args:
        coords (str):
    Returns:
        (str): The incremented row coordinates.
    """
    column = ''.join([i for i in coords if not i.isdigit()])
    seq_type = type(coords)
    row = int(seq_type().join(filter(seq_type.isdigit, coords))) + 1
    return column + str(row)


def import_worksheet(
        filename: str,
        sheetname: str,
        range_start: Optional[str] = 'A1',
    ) -> List[dict]:
    """Read the data from a given worksheet using xlwings.
    Args:
        filename (str): The name of the Excel workbook to read.
        sheetname (str): The name of the worksheet to import.
        range_start (str): Optional starting cell.
    Returns:
        list(dict): A list of dictionaries.
    """
    app = xlwings.App(visible=False)
    book = xlwings.Book(filename)
    sheet = book.sheets(sheetname)
    excel_data = sheet.range(range_start).expand('table').value
    keys = [snake_case(key) for key in excel_data[0]]
    data = [dict(zip(keys, values)) for values in excel_data[1:]]
    book.close()
    app.quit()
    return data


@xlwings.sub
def import_worksheet_data(model_type: str):
    """A function called from Excel to import data by IDs in the worksheet
    from Firestore into the Excel workbook.
    Args:
        model_type (str): The data model at hand.
    """

    # Initialize the workbook.
    worksheet, config = initialize_workbook()
    status_cell = config['status_cell']
    show_status_message(
        worksheet,
        coords=status_cell,
        message='Importing %s data...' % model_type,
        background=config['success_color'],
    )

    # Read the IDs.
    id_cell = increment_row(config['table_cell'])
    ids = worksheet.range(id_cell).options(expand='down', ndim=1).value

    # Get the worksheet columns.
    columns = worksheet.range(config['table_cell']).options(expand='right', ndim=1).value
    columns = [snake_case(x) for x in columns]

    # Get data using model type and ID through the API.
    response = get_data_model(worksheet, config, model_type, ids)
    if response.status_code != 200:
        show_status_message(
            worksheet,
            coords=status_cell,
            message='Error importing data.',
            background=config['error_color']
        )
        return

    # Format the values.
    try:
        data = response.json()['data']
        items = format_values(data, columns)
    except TypeError:
        show_status_message(
            worksheet,
            coords=status_cell,
            message='No data found.',
            background=config['error_color']
        )
        return

    # Insert all rows at the same time.
    worksheet.range(id_cell).value = items

    # Show success status message.
    show_status_message(
        worksheet,
        coords=status_cell,
        message='Imported %i %s.' % (len(ids), model_type),
    )


def initialize_workbook() -> Tuple:
    """Initialize an xlwings workbook and return the config worksheet data.
    Returns:
        (Worksheet): Returns the active worksheet.
        (dict): Returns a dictionary of user configuration.
    """
    book = xlwings.Book.caller()
    worksheet = book.sheets.active
    config_sheet = book.sheets['cannlytics.conf']
    config = get_data_block(config_sheet, 'A1', expand='table')
    return worksheet, config


def read_table_data(worksheet: Any, cell: str) -> dict:
    """Read table data from a worksheet and clean the columns and data.
    Args:
        worksheet (Worksheet): A worksheet containing the table data.
        cell (str): The cell range of the table.
    Returns:
        (dict): Returns the data in the worksheet table as a dictionary.
    """
    table = worksheet.range(cell)
    data = table.options(DataFrame, index=False, expand='table').value
    data.columns = list(map(snake_case, data.columns))
    for column in data.columns:
        if column.endswith('_at'):
            try:
                data[column] = to_datetime(data[column]).dt.strftime('%Y-%m-%dT%H:%M%:%SZ')
            except:
                pass
    return data.fillna('')


@xlwings.sub
def upload_worksheet_data(model_type):
    """A function called from Excel to import data by IDs in the worksheet
    from Firestore into the Excel workbook."""

    # Initialize the workbook.
    worksheet, config = initialize_workbook()
    status_cell = config['status_cell']
    show_status_message(
        worksheet,
        coords=status_cell,
        message='Uploading %s data...' % model_type,
        background=config['success_color'],
    )

    # Read the table data, cleaning the column names.
    data = read_table_data(worksheet, config['table_cell'])

    # Determine the model type and the organization.
    org_id = worksheet.range(config['org_id_cell']).value
    model_singular = data.columns[0].replace('_id', '')
    if not org_id:
        show_status_message(
            worksheet,
            coords=status_cell,
            message='Organization ID required.',
            background=config['error_color']
        )
        return

    # Get Cannlytics API key from .env using env_path in config.
    env_variables = dotenv_values(config['env_path'])
    api_key = env_variables['CANNLYTICS_API_KEY']

    # TODO: Split this into a separate function.
    # Upload data using model type, ID, and data through the API.
    headers = {
        'Authorization': 'Bearer %s' % api_key,
        'Content-type': 'application/json',
    }
    base = config['api_url']
    for _, row in data.iterrows():
        # Optional: Clean based on data model fields.
        json = row.to_dict()
        doc_id = json[f'{model_singular}_id']
        if not doc_id:
            continue
        url = f'{base}/{model_type}?organization_id={org_id}'
        response = requests.post(url, json=json, headers=headers)
        if response.status_code == 200:
            show_status_message(
                worksheet,
                coords=status_cell,
                message='Uploaded %s' % doc_id,
            )
        else:
            show_status_message(
                worksheet,
                coords=status_cell,
                message='Error uploading %s. Check your organization, internet connection and API key.' % doc_id, # pylint:disable=line-too-long
                background=config['error_color']
            )
            return

    # Show success status message.
    show_status_message(
        worksheet,
        coords=status_cell,
        message='Uploaded %i %s.' % (len(data), model_type),
    )


def show_status_message(sheet, coords, message, background=None, color=None):
    """Show a status message in an Excel spreadsheet.
    Args:
        sheet (Sheet): The sheet where the status message will be written.
        coords (str): The location of the status message.
        message (str): A status message to write to Excel.
        background (tuple): Optional background color.
        color (tuple): Optional font color.
    """
    sheet.range(coords).value = message
    if background:
        sheet.range(coords).color = literal_eval(background)
    if color:
        sheet.range(coords).api.Font.Color = rgb_to_int(literal_eval(color))


# openpyxl and pandas methods for reading data.

# def get_worksheet_headers(sheet):
#     """Get the headres of a worksheet.
#     Args:
#         sheet (Worksheet): An openpyx; Excel file object.
#     Returns:
#         headers (list): A list of header strings.
#     """
#     headers = []
#     for cell in sheet[1]:
#         headers.append(snake_case(cell.value))
#     return headers


# def get_worksheet_data(sheet, headers):
#     """Get the data of a worksheet.
#     Args:
#         sheet (Worksheet): An openpyx; Excel file object.
#         headres (list): A list of headers to map the values.
#     Returns:
#         list(dict): A list of dictionaries.
#     """
#     data = []
#     for row in sheet.iter_rows(min_row=2):
#         values = {}
#         for key, cell in zip(headers, row):
#             values[key] = cell.value
#         data.append(values)
#     return data


# def read_worksheet(path, filename='Upload'):
#     """Read the imported data, iterating over the rows and
#     getting value from each cell in row.
#     Args:
#         path (str or InMemoryFile): An Excel workbook to read.
#         filename (str): The name of the worksheet to upload.
#     Returns:
#         (DataFrame): A Pandas DataFrame of the results.
#     """
#     # try:
#     #     workbook = openpyxl.load_workbook(path, data_only=True)
#     #     sheet = workbook.get_sheet_by_name(filename)
#     #     headers = get_worksheet_headers(sheet)
#     #     return pd.DataFrame(get_worksheet_data(sheet, headers))
#     # except:
#     data = pd.read_csv(path)
#     data.columns = [snake_case(x) for x in data.columns]
#     return data



"""
Measurement Management | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 11/2/2021  
Updated: 11/2/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Manage your laboratory measurements.
"""

def record_measurement(config):
    """Record a laboratory measurement, e.g. a measured sample weight."""
    # pics = []
    # return pics
    return NotImplementedError


if __name__ == '__main__':
    
    # TODO: Test recording a measurement.
    record_measurement({})




"""
Statistics | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 6/23/2021  
Updated: 6/23/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Calculate statistics on your laboratory performance.
"""


def calculate_turnaround_time():
    """Calculate the average turnaround time of a group of samples or
    projects."""
    return NotImplementedError


def calculate_samples_tested():
    """Calculate the total number of samples tested."""
    return NotImplementedError


def calculate_organizations_served():
    """Calculate the total number of organizations served."""
    return NotImplementedError


def calculate_projects_completed():
    """Calculate the total number of projects completed."""
    return NotImplementedError


def calculate_avg_failure_rate():
    """Calculate the avergage failure rate for an analyte in a group of samples."""
    return NotImplementedError


def calculate_avg_result():
    """Calculate the avergage result for an analyte in a group of samples."""
    return NotImplementedError



"""
Sample Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Sample data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model


@dataclass
class Sample(Model):
    """A sample sent by a client organization to a lab organization
    for the lab to perform analyses on the sample and return results
    and a certificate to the client. Measurements will be made for the
    sample by analysts and/or scientific instruments, calculations will
    be applied to the sample's results, the sample's results are reviewed,
    a certificate is created for the sample's analyes, a CoA, at which
    point the lab can make the results and certificate accessible to the
    client."""
    _collection = 'organizations/%s/samples'
    batch_id: str = ''
    created_at: datetime = None
    created_by: str = ''
    photo_ref: str = ''
    photo_url: str = ''
    project_id: str = ''
    updated_at: datetime = None
    updated_by: str = ''
    notes: str = ''
